from __future__ import annotations

import itertools
from collections.abc import Callable, Collection, Generator, Iterable, Iterator, Mapping
from functools import partial
from typing import TYPE_CHECKING, Any, Concatenate, Self, overload

import cytoolz as cz
import more_itertools as mit

from .._core import IterWrapper

if TYPE_CHECKING:
    from ._main import Iter


class BaseMap[T](IterWrapper[T]):
    def for_each[**P](
        self,
        func: Callable[Concatenate[T, P], Any],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> Self:
        """
        Apply a function to each element in the iterable.

        Args:
            func: Function to apply to each element.
            args: Positional arguments for the function.
            kwargs: Keyword arguments for the function.

        Can be used for side effects such as printing or logging.
        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_([1, 2, 3]).for_each(lambda x: print(x)).collect().unwrap()
        1
        2
        3
        []

        ```
        """
        for v in self.unwrap():
            func(v, *args, **kwargs)
        return self

    def map[R](self, func: Callable[[T], R]) -> Iter[R]:
        """
        Map each element through func and return a Iter of results.

        Args:
            func: Function to apply to each element.

        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_([1, 2]).map(lambda x: x + 1).into(list)
        [2, 3]

        ```
        """
        return self.apply(partial(map, func))

    @overload
    def flat_map[U, R](
        self: IterWrapper[Iterable[Iterable[Iterable[U]]]],
        func: Callable[[T], Iterable[Iterable[R]]],
    ) -> Iter[Iterable[Iterable[R]]]: ...
    @overload
    def flat_map[U, R](
        self: IterWrapper[Iterable[Iterable[U]]], func: Callable[[T], Iterable[R]]
    ) -> Iter[Iterable[R]]: ...
    @overload
    def flat_map[U, R](
        self: IterWrapper[Iterable[U]], func: Callable[[T], R]
    ) -> Iter[R]: ...
    def flat_map[U: Iterable[Any], R](
        self: IterWrapper[U], func: Callable[[T], R]
    ) -> Iter[Any]:
        """
        Map each element through func and flatten the result by one level.

        Args:
            func: Function to apply to each element.

        ```python
        >>> import pyochain as pc
        >>> data = [[1, 2], [3, 4]]
        >>> pc.Iter.from_(data).flat_map(lambda x: x + 10).into(list)
        [11, 12, 13, 14]

        ```
        """

        def _flat_map(data: Iterable[U]) -> map[R]:
            return map(func, itertools.chain.from_iterable(data))

        return self.apply(_flat_map)

    def map_star[U: Iterable[Any], R](
        self: IterWrapper[U], func: Callable[..., R]
    ) -> Iter[R]:
        """
        Applies a function to each element, where each element is an iterable.

        Args:
            func: Function to apply to unpacked elements.

        Unlike `.map()`, which passes each element as a single argument, `.starmap()` unpacks each element into positional arguments for the function.

        In short, for each `element` in the sequence, it computes `func(*element)`.
        ```python
        >>> import pyochain as pc
        >>> def make_sku(color, size):
        ...     return f"{color}-{size}"
        >>> data = pc.Seq(["blue", "red"])
        >>> data.iter().product(["S", "M"]).map_star(make_sku).into(list)
        ['blue-S', 'blue-M', 'red-S', 'red-M']

        ```
        This is equivalent to:
        ```python
        >>> data.iter().product(["S", "M"]).map(lambda x: make_sku(*x)).into(list)
        ['blue-S', 'blue-M', 'red-S', 'red-M']

        ```
        - Use map_star when the performance matters (it is faster).
        - Use map with unpacking when readability matters (the types can be inferred).
        """

        return self.apply(partial(itertools.starmap, func))

    def map_if[R](
        self,
        predicate: Callable[[T], bool],
        func: Callable[[T], R],
        func_else: Callable[[T], R] | None = None,
    ) -> Iter[R]:
        """
        Evaluate each item from iterable using pred.

        Args:
            predicate: Function to evaluate each item.
            func: Function to apply if predicate is True.
            func_else: Function to apply if predicate is False.

        - If the result is equivalent to True, transform the item with func and yield it.
        - Otherwise, transform the item with func_else and yield it.
        - Predicate, func, and func_else should each be functions that accept one argument.

        By default, func_else is the identity function.
        ```python
        >>> import pyochain as pc
        >>> from math import sqrt
        >>> iterable = pc.Iter.from_(range(-5, 5)).collect()
        >>> iterable.into(list)
        [-5, -4, -3, -2, -1, 0, 1, 2, 3, 4]
        >>> iterable.iter().map_if(lambda x: x > 3, lambda x: "toobig").into(list)
        [-5, -4, -3, -2, -1, 0, 1, 2, 3, 'toobig']
        >>> iterable.iter().map_if(
        ...     lambda x: x >= 0,
        ...     lambda x: f"{sqrt(x):.2f}",
        ...     lambda x: None,
        ... ).into(list)
        [None, None, None, None, None, '0.00', '1.00', '1.41', '1.73', '2.00']

        ```
        """
        return self.apply(mit.map_if, predicate, func, func_else=func_else)

    def map_except[R](
        self, func: Callable[[T], R], *exceptions: type[BaseException]
    ) -> Iter[R]:
        """
        Transform each item from iterable with function and yield the result, unless function raises one of the specified exceptions.

        Args:
            func: Function to apply to each item.
            exceptions: Exceptions to catch and ignore.

        The function is called to transform each item in iterable

        If an exception other than one given by exceptions is raised by function, it is raised like normal.
        ```python
        >>> import pyochain as pc
        >>> iterable = ["1", "2", "three", "4", None]
        >>> pc.Iter.from_(iterable).map_except(int, ValueError, TypeError).into(list)
        [1, 2, 4]

        ```
        """

        def _map_except(data: Iterable[T]) -> Iterator[R]:
            return mit.map_except(func, data, *exceptions)

        return self.apply(_map_except)

    def repeat(
        self, n: int, factory: Callable[[Iterable[T]], Collection[T]] = tuple
    ) -> Iter[Iterable[T]]:
        """
        Repeat the entire iterable n times (as elements) and return Iter.

        Args:
            n: Number of repetitions.
            factory: Factory to create the repeated collection (default: tuple).

        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_([1, 2]).repeat(2).collect().unwrap()
        [(1, 2), (1, 2)]
        >>> pc.Iter.from_([1, 2]).repeat(3, list).collect().unwrap()
        [[1, 2], [1, 2], [1, 2]]

        ```
        """

        def _repeat(data: Iterable[T]) -> Iterator[Iterable[T]]:
            return itertools.repeat(factory(data), n)

        return self.apply(_repeat)

    @overload
    def repeat_last(self, default: T) -> Iter[T]: ...
    @overload
    def repeat_last[U](self, default: U) -> Iter[T | U]: ...
    def repeat_last[U](self, default: U = None) -> Iter[T | U]:
        """
        After the iterable is exhausted, keep yielding its last element.

        **Warning** ⚠️
            This creates an infinite iterator.
            Be sure to use `Iter.take()` or `Iter.slice()` to limit the number of items taken.

        Args:
            default: Value to yield if the iterable is empty.

        Example:
        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_(range(3)).repeat_last().take(5).into(list)
        [0, 1, 2, 2, 2]

        If the iterable is empty, yield default forever:
        ```python
        >>> pc.Iter.from_(range(0)).repeat_last(42).take(5).into(list)
        [42, 42, 42, 42, 42]

        ```
        """
        return self.apply(mit.repeat_last, default)

    def ichunked(self, n: int) -> Iter[Iterator[T]]:
        """

        Break *iterable* into sub-iterables with *n* elements each.

        Args:
            n: Number of elements in each chunk.

        If the sub-iterables are read in order, the elements of *iterable*
        won't be stored in memory.

        If they are read out of order, :func:`itertools.tee` is used to cache
        elements as necessary.
        ```python
        >>> import pyochain as pc
        >>> all_chunks = pc.Iter.from_count().ichunked(4).unwrap()
        >>> c_1, c_2, c_3 = next(all_chunks), next(all_chunks), next(all_chunks)
        >>> list(c_2)  # c_1's elements have been cached; c_3's haven't been
        [4, 5, 6, 7]
        >>> list(c_1)
        [0, 1, 2, 3]
        >>> list(c_3)
        [8, 9, 10, 11]

        ```
        """
        return self.apply(mit.ichunked, n)

    @overload
    def flatten[U](
        self: IterWrapper[Iterable[Iterable[Iterable[U]]]],
    ) -> Iter[Iterable[Iterable[U]]]: ...
    @overload
    def flatten[U](self: IterWrapper[Iterable[Iterable[U]]]) -> Iter[Iterable[U]]: ...
    @overload
    def flatten[U](self: IterWrapper[Iterable[U]]) -> Iter[U]: ...
    def flatten(self: IterWrapper[Iterable[Any]]) -> Iter[Any]:
        """
        Flatten one level of nesting and return a new Iterable wrapper.

        This is a shortcut for `.apply(itertools.chain.from_iterable)`.
        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_([[1, 2], [3]]).flatten().into(list)
        [1, 2, 3]

        ```
        """
        return self.apply(itertools.chain.from_iterable)

    def pluck[U: Mapping[Any, Any]](
        self: IterWrapper[U], *keys: str | int
    ) -> Iter[Any]:
        """
        Get an element from each item in a sequence using a nested key path.

        Args:
            keys: Nested keys to extract values.

        ```python
        >>> import pyochain as pc
        >>> data = pc.Seq(
        ...     [
        ...         {"id": 1, "info": {"name": "Alice", "age": 30}},
        ...         {"id": 2, "info": {"name": "Bob", "age": 25}},
        ...     ]
        ... )
        >>> data.iter().pluck("info").into(list)
        [{'name': 'Alice', 'age': 30}, {'name': 'Bob', 'age': 25}]
        >>> data.iter().pluck("info", "name").into(list)
        ['Alice', 'Bob']

        ```
        Example: get the maximum age along with the corresponding id)
        ```python
        >>> data.iter().pluck("info", "age").zip(
        ...     data.iter().pluck("id").into(list)
        ... ).max()
        (30, 1)

        ```
        """

        getter = partial(cz.dicttoolz.get_in, keys)
        return self.apply(partial(map, getter))

    def round[U: float | int](
        self: IterWrapper[U], ndigits: int | None = None
    ) -> Iter[float]:
        """
        Round each element in the iterable to the given number of decimal places and return Iter.

        Args:
            ndigits: Number of decimal places to round to.

        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_([1.2345, 2.3456, 3.4567]).round(2).into(list)
        [1.23, 2.35, 3.46]

        ```
        """

        def _round(data: Iterable[U]) -> Generator[float | int, None, None]:
            return (round(x, ndigits) for x in data)

        return self.apply(_round)



from __future__ import annotations

import functools
import statistics
from collections.abc import Callable, Iterable
from typing import TYPE_CHECKING, Literal, NamedTuple

import cytoolz as cz
import more_itertools as mit

from .._core import IterWrapper

if TYPE_CHECKING:
    from ._main import Iter


class Unzipped[T, V](NamedTuple):
    first: Iter[T]
    second: Iter[V]


class BaseAgg[T](IterWrapper[T]):
    def unzip[U, V](self: IterWrapper[tuple[U, V]]) -> Unzipped[U, V]:
        """
        Converts an iterator of pairs into a pair of iterators.

        `Iter.unzip()` consumes the iterator of pairs.

        Returns an Unzipped NamedTuple, containing two iterators:

        - one from the left elements of the pairs
        - one from the right elements.
        This function is, in some sense, the opposite of zip.
        ```python
        >>> import pyochain as pc
        >>> data = [(1, "a"), (2, "b"), (3, "c")]
        >>> unzipped = pc.Iter.from_(data).unzip()
        >>> unzipped.first.into(list)
        [1, 2, 3]
        >>> unzipped.second.into(list)
        ['a', 'b', 'c']

        ```
        """
        from ._main import Iter

        def _unzip(data: Iterable[tuple[U, V]]) -> Unzipped[U, V]:
            d: list[tuple[U, V]] = list(data)
            return Unzipped(Iter(x[0] for x in d), Iter(x[1] for x in d))

        return self.into(_unzip)

    def reduce(self, func: Callable[[T, T], T]) -> T:
        """
        Apply a function of two arguments cumulatively to the items of an iterable, from left to right.

        Args:
            func: Function to apply cumulatively to the items of the iterable.

        This effectively reduces the iterable to a single value.

        If initial is present, it is placed before the items of the iterable in the calculation.

        It then serves as a default when the iterable is empty.
        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_([1, 2, 3]).reduce(lambda a, b: a + b)
        6

        ```
        """
        return self.into(functools.partial(functools.reduce, func))

    def combination_index(self, r: Iterable[T]) -> int:
        """
        Equivalent to list(combinations(iterable, r)).index(element).

        Args:
            r: The combination to find the index of.

        The subsequences of iterable that are of length r can be ordered lexicographically.

        combination_index computes the index of the first element, without computing the previous combinations.

        ValueError will be raised if the given element isn't one of the combinations of iterable.
        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_("abcdefg").combination_index("adf")
        10

        ```
        """
        return self.into(functools.partial(mit.combination_index, r))

    def first(self) -> T:
        """
        Return the first element.
        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_([9]).first()
        9

        ```
        """
        return self.into(cz.itertoolz.first)

    def second(self) -> T:
        """
        Return the second element.
        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_([9, 8]).second()
        8

        ```
        """
        return self.into(cz.itertoolz.second)

    def last(self) -> T:
        """
        Return the last element.
        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_([7, 8, 9]).last()
        9

        ```
        """
        return self.into(cz.itertoolz.last)

    def count(self) -> int:
        """
        Return the length of the sequence.
        Like the builtin len but works on lazy sequences.
        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_([1, 2]).count()
        2

        ```
        """
        return self.into(cz.itertoolz.count)

    def item(self, index: int) -> T:
        """
        Return item at index.

        Args:
            index: The index of the item to retrieve.

        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_([10, 20]).item(1)
        20

        ```
        """
        return self.into(functools.partial(cz.itertoolz.nth, index))

    def argmax[U](self, key: Callable[[T], U] | None = None) -> int:
        """
        Index of the first occurrence of a maximum value in an iterable.

        Args:
            key: Optional function to determine the value for comparison.

        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_("abcdefghabcd").argmax()
        7
        >>> pc.Iter.from_([0, 1, 2, 3, 3, 2, 1, 0]).argmax()
        3

        ```
        For example, identify the best machine learning model:
        ```python
        >>> models = pc.Iter.from_(["svm", "random forest", "knn", "naïve bayes"])
        >>> accuracy = pc.Seq([68, 61, 84, 72])
        >>> # Most accurate model
        >>> models.item(accuracy.argmax())
        'knn'
        >>>
        >>> # Best accuracy
        >>> accuracy.into(max)
        84

        ```
        """
        return self.into(mit.argmax, key=key)

    def argmin[U](self, key: Callable[[T], U] | None = None) -> int:
        """
        Index of the first occurrence of a minimum value in an iterable.

        Args:
            key: Optional function to determine the value for comparison.

        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_("efghabcdijkl").argmin()
        4
        >>> pc.Iter.from_([3, 2, 1, 0, 4, 2, 1, 0]).argmin()
        3

        ```

        For example, look up a label corresponding to the position of a value that minimizes a cost function:
        ```python
        >>> def cost(x):
        ...     "Days for a wound to heal given a subject's age."
        ...     return x**2 - 20 * x + 150
        >>> labels = pc.Iter.from_(["homer", "marge", "bart", "lisa", "maggie"])
        >>> ages = pc.Seq([35, 30, 10, 9, 1])
        >>> # Fastest healing family member
        >>> labels.item(ages.argmin(key=cost))
        'bart'
        >>> # Age with fastest healing
        >>> ages.into(min, key=cost)
        10

        ```
        """
        return self.into(mit.argmin, key=key)

    def sum[U: int | float](self: IterWrapper[U]) -> U | Literal[0]:
        """
        Return the sum of the sequence.
        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_([1, 2, 3]).sum()
        6

        ```
        """
        return self.into(sum)

    def min[U: int | float](self: IterWrapper[U]) -> U:
        """
        Return the minimum of the sequence.
        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_([3, 1, 2]).min()
        1

        ```
        """
        return self.into(min)

    def max[U: int | float](self: IterWrapper[U]) -> U:
        """
        Return the maximum of the sequence.
        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_([3, 1, 2]).max()
        3

        ```
        """
        return self.into(max)

    def mean[U: int | float](self: IterWrapper[U]) -> float:
        """
        Return the mean of the sequence.
        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_([1, 2, 3]).mean()
        2

        ```
        """
        return self.into(statistics.mean)

    def median[U: int | float](self: IterWrapper[U]) -> float:
        """
        Return the median of the sequence.
        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_([1, 3, 2]).median()
        2

        ```
        """
        return self.into(statistics.median)

    def mode[U: int | float](self: IterWrapper[U]) -> U:
        """
        Return the mode of the sequence.
        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_([1, 2, 2, 3]).mode()
        2

        ```
        """
        return self.into(statistics.mode)

    def stdev[U: int | float](
        self: IterWrapper[U],
    ) -> float:
        """
        Return the standard deviation of the sequence.
        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_([1, 2, 3]).stdev()
        1.0

        ```
        """
        return self.into(statistics.stdev)

    def variance[U: int | float](
        self: IterWrapper[U],
    ) -> float:
        """
        Return the variance of the sequence.
        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_([1, 2, 3, 7, 8]).variance()
        9.7

        ```
        """
        return self.into(statistics.variance)



from __future__ import annotations

from collections.abc import Callable, Iterable
from typing import overload

import cytoolz as cz
import more_itertools as mit

from .._core import IterWrapper


class BaseBool[T](IterWrapper[T]):
    def all(self, predicate: Callable[[T], bool] = lambda x: bool(x)) -> bool:
        """
        Tests if every element of the iterator matches a predicate.

        `Iter.all()` takes a closure that returns true or false.

        It applies this closure to each element of the iterator, and if they all return true, then so does `Iter.all()`.

        If any of them return false, it returns false.

        An empty iterator returns true.
        Args:
            predicate: Function to evaluate each item. Defaults to checking truthiness.
        Example:
        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_([1, True]).all()
        True
        >>> pc.Iter.from_([]).all()
        True
        >>> pc.Iter.from_([1, 0]).all()
        False
        >>> def is_even(x: int) -> bool:
        ...     return x % 2 == 0
        >>> pc.Iter.from_([2, 4, 6]).all(is_even)
        True

        ```
        """

        def _all(data: Iterable[T]) -> bool:
            return all(predicate(x) for x in data)

        return self.into(_all)

    def any(self, predicate: Callable[[T], bool] = lambda x: bool(x)) -> bool:
        """
        Tests if any element of the iterator matches a predicate.


        `Iter.any()` takes a closure that returns true or false.

        It applies this closure to each element of the iterator, and if any of them return true, then so does `Iter.any()`.

        If they all return false, it returns false.

        An empty iterator returns false.
        Args:
            predicate: Function to evaluate each item. Defaults to checking truthiness.
        Example:
        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_([0, 1]).any()
        True
        >>> pc.Iter.from_(range(0)).any()
        False
        >>> def is_even(x: int) -> bool:
        ...     return x % 2 == 0
        >>> pc.Iter.from_([1, 3, 4]).any(is_even)
        True

        ```
        """

        def _any(data: Iterable[T]) -> bool:
            return any(predicate(x) for x in data)

        return self.into(_any)

    def is_distinct(self) -> bool:
        """
        Return True if all items are distinct.
        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_([1, 2]).is_distinct()
        True

        ```
        """
        return self.into(cz.itertoolz.isdistinct)

    def all_equal[U](self, key: Callable[[T], U] | None = None) -> bool:
        """
        Return True if all items are equal.

        Args:
            key: Function to transform items before comparison. Defaults to None.
        Example:
        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_([1, 1, 1]).all_equal()
        True

        ```
        A function that accepts a single argument and returns a transformed version of each input item can be specified with key:
        ```python
        >>> pc.Iter.from_("AaaA").all_equal(key=str.casefold)
        True
        >>> pc.Iter.from_([1, 2, 3]).all_equal(key=lambda x: x < 10)
        True

        ```
        """
        return self.into(mit.all_equal, key=key)

    def all_unique[U](self, key: Callable[[T], U] | None = None) -> bool:
        """
        Returns True if all the elements of iterable are unique.

        Args:
            key: Function to transform items before comparison. Defaults to None.
        Example:
        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_("ABCB").all_unique()
        False

        ```
        If a key function is specified, it will be used to make comparisons.
        ```python
        >>> pc.Iter.from_("ABCb").all_unique()
        True
        >>> pc.Iter.from_("ABCb").all_unique(str.lower)
        False

        ```
        The function returns as soon as the first non-unique element is encountered.

        Iterables with a mix of hashable and unhashable items can be used, but the function will be slower for unhashable items

        """
        return self.into(mit.all_unique, key=key)

    def is_sorted[U](
        self,
        key: Callable[[T], U] | None = None,
        reverse: bool = False,
        strict: bool = False,
    ) -> bool:
        """
        Returns True if the items of iterable are in sorted order.

        Args:
            key: Function to transform items before comparison. Defaults to None.
            reverse: Whether to check for descending order. Defaults to False.
            strict: Whether to enforce strict sorting (no equal elements). Defaults to False.
        Example:
        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_(["1", "2", "3", "4", "5"]).is_sorted(key=int)
        True
        >>> pc.Iter.from_([5, 4, 3, 1, 2]).is_sorted(reverse=True)
        False

        If strict, tests for strict sorting, that is, returns False if equal elements are found:
        ```python
        >>> pc.Iter.from_([1, 2, 2]).is_sorted()
        True
        >>> pc.Iter.from_([1, 2, 2]).is_sorted(strict=True)
        False

        ```

        The function returns False after encountering the first out-of-order item.

        This means it may produce results that differ from the built-in sorted function for objects with unusual comparison dynamics (like math.nan).

        If there are no out-of-order items, the iterable is exhausted.
        """
        return self.into(mit.is_sorted, key=key, reverse=reverse, strict=strict)

    @overload
    def find(
        self, default: None = None, predicate: Callable[[T], bool] | None = ...
    ) -> T | None: ...
    @overload
    def find(self, default: T, predicate: Callable[[T], bool] | None = ...) -> T: ...

    def find[U](
        self, default: U = None, predicate: Callable[[T], bool] | None = None
    ) -> U | T:
        """
        Searches for an element of an iterator that satisfies a `predicate`, by:

        - Taking a closure that returns true or false as `predicate` (optional).
        - Using the identity function if no `predicate` is provided.
        - Applying this closure to each element of the iterator.
        - Returning the first element that satisfies the `predicate`.

        If all the elements return false, `Iter.find()` returns the default value.
        Args:
            default: Value to return if no element satisfies the predicate. Defaults to None.
            predicate: Function to evaluate each item. Defaults to checking truthiness.
        Example:
        ```python
        >>> import pyochain as pc
        >>> def gt_five(x: int) -> bool:
        ...     return x > 5
        >>>
        >>> def gt_nine(x: int) -> bool:
        ...     return x > 9
        >>>
        >>> pc.Iter.from_(range(10)).find()
        1
        >>> pc.Iter.from_(range(10)).find(predicate=gt_five)
        6
        >>> pc.Iter.from_(range(10)).find(default="missing", predicate=gt_nine)
        'missing'

        ```
        """
        return self.into(mit.first_true, default, predicate)



from __future__ import annotations

import itertools
from collections.abc import Callable
from functools import partial
from typing import TYPE_CHECKING, Literal, overload

import cytoolz as cz
import more_itertools as mit

from .._core import IterWrapper

if TYPE_CHECKING:
    from ._main import Iter


class BaseTuples[T](IterWrapper[T]):
    def enumerate(self) -> Iter[tuple[int, T]]:
        """
        Return a Iter of (index, value) pairs.
        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_(["a", "b"]).enumerate().into(list)
        [(0, 'a'), (1, 'b')]

        ```
        """
        return self.apply(enumerate)

    @overload
    def combinations(self, r: Literal[2]) -> Iter[tuple[T, T]]: ...
    @overload
    def combinations(self, r: Literal[3]) -> Iter[tuple[T, T, T]]: ...
    @overload
    def combinations(self, r: Literal[4]) -> Iter[tuple[T, T, T, T]]: ...
    @overload
    def combinations(self, r: Literal[5]) -> Iter[tuple[T, T, T, T, T]]: ...
    def combinations(self, r: int) -> Iter[tuple[T, ...]]:
        """
        Return all combinations of length r.

        Args:
            r: Length of each combination.

        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_([1, 2, 3]).combinations(2).into(list)
        [(1, 2), (1, 3), (2, 3)]

        ```
        """
        return self.apply(itertools.combinations, r)

    @overload
    def permutations(self, r: Literal[2]) -> Iter[tuple[T, T]]: ...
    @overload
    def permutations(self, r: Literal[3]) -> Iter[tuple[T, T, T]]: ...
    @overload
    def permutations(self, r: Literal[4]) -> Iter[tuple[T, T, T, T]]: ...
    @overload
    def permutations(self, r: Literal[5]) -> Iter[tuple[T, T, T, T, T]]: ...
    def permutations(self, r: int | None = None) -> Iter[tuple[T, ...]]:
        """
        Return all permutations of length r.

        Args:
            r: Length of each permutation. Defaults to the length of the iterable.

        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_([1, 2, 3]).permutations(2).into(list)
        [(1, 2), (1, 3), (2, 1), (2, 3), (3, 1), (3, 2)]

        ```
        """
        return self.apply(itertools.permutations, r)

    @overload
    def combinations_with_replacement(self, r: Literal[2]) -> Iter[tuple[T, T]]: ...
    @overload
    def combinations_with_replacement(self, r: Literal[3]) -> Iter[tuple[T, T, T]]: ...
    @overload
    def combinations_with_replacement(
        self, r: Literal[4]
    ) -> Iter[tuple[T, T, T, T]]: ...
    @overload
    def combinations_with_replacement(
        self, r: Literal[5]
    ) -> Iter[tuple[T, T, T, T, T]]: ...
    def combinations_with_replacement(self, r: int) -> Iter[tuple[T, ...]]:
        """
        Return all combinations with replacement of length r.

        Args:
            r: Length of each combination.

        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_([1, 2, 3]).combinations_with_replacement(2).into(list)
        [(1, 1), (1, 2), (1, 3), (2, 2), (2, 3), (3, 3)]

        ```
        """
        return self.apply(itertools.combinations_with_replacement, r)

    def pairwise(self) -> Iter[tuple[T, T]]:
        """
        Return an iterator over pairs of consecutive elements.
        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_([1, 2, 3]).pairwise().into(list)
        [(1, 2), (2, 3)]

        ```
        """
        return self.apply(itertools.pairwise)

    @overload
    def map_juxt[R1, R2](
        self,
        func1: Callable[[T], R1],
        func2: Callable[[T], R2],
        /,
    ) -> Iter[tuple[R1, R2]]: ...
    @overload
    def map_juxt[R1, R2, R3](
        self,
        func1: Callable[[T], R1],
        func2: Callable[[T], R2],
        func3: Callable[[T], R3],
        /,
    ) -> Iter[tuple[R1, R2, R3]]: ...
    @overload
    def map_juxt[R1, R2, R3, R4](
        self,
        func1: Callable[[T], R1],
        func2: Callable[[T], R2],
        func3: Callable[[T], R3],
        func4: Callable[[T], R4],
        /,
    ) -> Iter[tuple[R1, R2, R3, R4]]: ...
    def map_juxt(self, *funcs: Callable[[T], object]) -> Iter[tuple[object, ...]]:
        """
        Apply several functions to each item.

        Returns a new Iter where each item is a tuple of the results of applying each function to the original item.
        ```python
        >>> import pyochain as pc
        >>> def is_even(n: int) -> bool:
        ...     return n % 2 == 0
        >>> def is_positive(n: int) -> bool:
        ...     return n > 0
        >>>
        >>> pc.Iter.from_([1, -2, 3]).map_juxt(is_even, is_positive).into(list)
        [(False, True), (True, False), (False, True)]

        ```
        """
        return self.apply(partial(map, cz.functoolz.juxt(*funcs)))

    def adjacent(
        self, predicate: Callable[[T], bool], distance: int = 1
    ) -> Iter[tuple[bool, T]]:
        """
        Return an iterable over (bool, item) tuples.

        Args:
            predicate: Function to determine if an item satisfies the condition.
            distance: Number of places to consider as adjacent. Defaults to 1.

        The output is a sequence of tuples where the item is drawn from iterable.

        The bool indicates whether that item satisfies the predicate or is adjacent to an item that does.

        For example, to find whether items are adjacent to a 3:
        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_(range(6)).adjacent(lambda x: x == 3).into(list)
        [(False, 0), (False, 1), (True, 2), (True, 3), (True, 4), (False, 5)]

        ```
        Set distance to change what counts as adjacent.
        For example, to find whether items are two places away from a 3:
        ```python
        >>> pc.Iter.from_(range(6)).adjacent(lambda x: x == 3, distance=2).into(list)
        [(False, 0), (True, 1), (True, 2), (True, 3), (True, 4), (True, 5)]

        ```

        This is useful for contextualizing the results of a search function.

        For example, a code comparison tool might want to identify lines that have changed, but also surrounding lines to give the viewer of the diff context.

        The predicate function will only be called once for each item in the iterable.

        See also groupby_transform, which can be used with this function to group ranges of items with the same bool value.
        """
        return self.apply(partial(mit.adjacent, predicate, distance=distance))

    def classify_unique(self) -> Iter[tuple[T, bool, bool]]:
        """
        Classify each element in terms of its uniqueness.\n
        For each element in the input iterable, return a 3-tuple consisting of:

        - The element itself
        - False if the element is equal to the one preceding it in the input, True otherwise (i.e. the equivalent of unique_justseen)
        - False if this element has been seen anywhere in the input before, True otherwise (i.e. the equivalent of unique_everseen)
        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_("otto").classify_unique().into(list)
        ... # doctest: +NORMALIZE_WHITESPACE
        [('o', True,  True),
        ('t', True,  True),
        ('t', False, False),
        ('o', True,  False)]

        ```

        This function is analogous to unique_everseen and is subject to the same performance considerations.
        """
        return self.apply(mit.classify_unique)



from __future__ import annotations

from collections.abc import Callable, Iterable
from dataclasses import dataclass
from typing import Any, Self, TypeGuard

import cytoolz as cz

from .._core import Pipeable


@dataclass(slots=True)
class Expr(Pipeable):
    """
    Represents an expression in the pipeline.

    An Expr encapsulates a sequence of operations to be applied to keys on a python dict.

    Each Expr instance maintains:
    - A list of tokens representing the keys to access in the dict (the first being the input given to the `key` function),
    - A tuple of operations to apply to the accessed data
    - An alias for the expression (default to the last token).
    """

    __tokens__: list[str]
    __ops__: tuple[Callable[[object], object], ...]
    _alias: str

    def __repr__(self) -> str:
        parts: list[str] = []
        s_parts: list[str] = []
        for t in self.__tokens__:
            parts.append(f"field({t!r})")
            if s_parts:
                s_parts.append(".")
            s_parts.append(str(t))
        symbolic = ".".join(parts) if parts else "<root>"
        lowered = "".join(s_parts) or "<root>"
        base = f"Expr({symbolic} -> {lowered})"
        return f"{base}.alias({self._alias!r})"

    def _to_expr(self, op: Callable[[Any], Any]) -> Self:
        return self.__class__(
            self.__tokens__,
            self.__ops__ + (op,),
            self._alias,
        )

    def key(self, name: str) -> Self:
        """
        Add a key to the expression.

        Allow to access nested keys in a dict.

        Args:
            name: The key to access.
        Example:
        ```python
        >>> import pyochain as pc
        >>> expr = pc.key("a").key("b").key("c")
        >>> expr.__tokens__
        ['a', 'b', 'c']
        >>> data = {"a": {"b": {"c": 42}}}
        >>> pc.Dict(data).select(expr).unwrap()
        {'c': 42}

        ```
        """
        return self.__class__(
            self.__tokens__ + [name],
            self.__ops__,
            name,
        )

    def alias(self, name: str) -> Self:
        return self.__class__(self.__tokens__, self.__ops__, name)

    @property
    def name(self) -> str:
        return self._alias

    def apply(self, fn: Callable[[Any], Any]) -> Self:
        """
        Applies the given function fn to the data within the current Expr instance
        """

        def _apply(data: Any) -> Any:
            return fn(data)

        return self._to_expr(_apply)


def key(name: str) -> Expr:
    """Create an Expr that accesses the given key."""
    return Expr([name], (), name)


def _expr_identity(obj: Any) -> TypeGuard[Expr]:
    return hasattr(obj, "__tokens__")


type IntoExpr = Expr | str


def compute_exprs(
    exprs: Iterable[IntoExpr], data_in: dict[str, Any], data_out: dict[str, Any]
) -> dict[str, Any]:
    for e in exprs:
        if not _expr_identity(e):
            e = key(e)  # type: ignore
        current: object = cz.dicttoolz.get_in(e.__tokens__, data_in)
        for op in e.__ops__:
            current = op(current)
        data_out[e.name] = current
    return data_out



from __future__ import annotations

import itertools
from collections.abc import Callable, Generator, Iterable
from typing import TYPE_CHECKING, Any

import more_itertools as mit

from .._core import IterWrapper

if TYPE_CHECKING:
    from ._main import Iter


class BaseList[T](IterWrapper[T]):
    def implode(self) -> Iter[list[T]]:
        """
        Wrap each element in the iterable into a list.

        Syntactic sugar for `Iter.map(lambda x: [x])`.
        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_(range(5)).implode().into(list)
        [[0], [1], [2], [3], [4]]

        ```
        """

        def _implode(data: Iterable[T]) -> Generator[list[T], None, None]:
            return ([x] for x in data)

        return self.apply(_implode)

    def split_at(
        self,
        pred: Callable[[T], bool],
        maxsplit: int = -1,
        keep_separator: bool = False,
    ) -> Iter[list[T]]:
        """
        Yield lists of items from iterable, where each list is delimited by an item where callable pred returns True.

        Args:
            pred: Function to determine the split points.
            maxsplit: Maximum number of splits to perform. Defaults to -1 (no limit).
            keep_separator: Whether to include the separator in the output. Defaults to False.
        Example:
        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_("abcdcba").split_at(lambda x: x == "b").into(list)
        [['a'], ['c', 'd', 'c'], ['a']]
        >>> pc.Iter.from_(range(10)).split_at(lambda n: n % 2 == 1).into(list)
        [[0], [2], [4], [6], [8], []]

        At most *maxsplit* splits are done.

        If *maxsplit* is not specified or -1, then there is no limit on the number of splits:
        ```python
        >>> pc.Iter.from_(range(10)).split_at(lambda n: n % 2 == 1, maxsplit=2).into(
        ...     list
        ... )
        [[0], [2], [4, 5, 6, 7, 8, 9]]

        ```
        By default, the delimiting items are not included in the output.

        To include them, set *keep_separator* to `True`.
        ```python
        >>> def cond(x: str) -> bool:
        ...     return x == "b"
        >>> pc.Iter.from_("abcdcba").split_at(cond, keep_separator=True).into(list)
        [['a'], ['b'], ['c', 'd', 'c'], ['b'], ['a']]

        ```
        """
        return self.apply(mit.split_at, pred, maxsplit, keep_separator)

    def split_after(
        self, predicate: Callable[[T], bool], max_split: int = -1
    ) -> Iter[list[T]]:
        """
        Yield lists of items from iterable, where each list ends with an item where callable pred returns True.

        Args:
            predicate: Function to determine the split points.
            max_split: Maximum number of splits to perform. Defaults to -1 (no limit).
        Example:
        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_("one1two2").split_after(str.isdigit).into(list)
        [['o', 'n', 'e', '1'], ['t', 'w', 'o', '2']]

        >>> def cond(n: int) -> bool:
        ...     return n % 3 == 0
        >>>
        >>> pc.Iter.from_(range(10)).split_after(cond).into(list)
        [[0], [1, 2, 3], [4, 5, 6], [7, 8, 9]]
        >>> pc.Iter.from_(range(10)).split_after(cond, max_split=2).into(list)
        [[0], [1, 2, 3], [4, 5, 6, 7, 8, 9]]

        ```
        """
        return self.apply(mit.split_after, predicate, max_split)

    def split_before(
        self, predicate: Callable[[T], bool], max_split: int = -1
    ) -> Iter[list[T]]:
        """
        Yield lists of items from iterable, where each list ends with an item where callable pred returns True.

        Args:
            predicate: Function to determine the split points.
            max_split: Maximum number of splits to perform. Defaults to -1 (no limit).
        Example:
        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_("abcdcba").split_before(lambda x: x == "b").into(list)
        [['a'], ['b', 'c', 'd', 'c'], ['b', 'a']]
        >>>
        >>> def cond(n: int) -> bool:
        ...     return n % 2 == 1
        >>>
        >>> pc.Iter.from_(range(10)).split_before(cond).into(list)
        [[0], [1, 2], [3, 4], [5, 6], [7, 8], [9]]

        ```
        At most *max_split* splits are done.

        If *max_split* is not specified or -1, then there is no limit on the number of splits:
        ```python
        >>> pc.Iter.from_(range(10)).split_before(cond, max_split=2).into(list)
        [[0], [1, 2], [3, 4, 5, 6, 7, 8, 9]]

        ```
        """
        return self.apply(mit.split_before, predicate, max_split)

    def split_into(self, sizes: Iterable[int | None]) -> Iter[list[T]]:
        """
        Yield a list of sequential items from iterable of length 'n' for each integer 'n' in sizes.

        Args:
            sizes: Iterable of integers specifying the sizes of each chunk. Use None for the remainder.
        Example:
        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_([1, 2, 3, 4, 5, 6]).split_into([1, 2, 3]).into(list)
        [[1], [2, 3], [4, 5, 6]]

        If the sum of sizes is smaller than the length of iterable, then the remaining items of iterable will not be returned.
        ```python
        >>> pc.Iter.from_([1, 2, 3, 4, 5, 6]).split_into([2, 3]).into(list)
        [[1, 2], [3, 4, 5]]

        ```

        If the sum of sizes is larger than the length of iterable:

        - fewer items will be returned in the iteration that overruns the iterable
        - further lists will be empty
        ```python
        >>> pc.Iter.from_([1, 2, 3, 4]).split_into([1, 2, 3, 4]).into(list)
        [[1], [2, 3], [4], []]

        ```

        When a None object is encountered in sizes, the returned list will contain items up to the end of iterable the same way that itertools.slice does:
        ```python
        >>> data = [1, 2, 3, 4, 5, 6, 7, 8, 9, 0]
        >>> pc.Iter.from_(data).split_into([2, 3, None]).into(list)
        [[1, 2], [3, 4, 5], [6, 7, 8, 9, 0]]

        ```

        split_into can be useful for grouping a series of items where the sizes of the groups are not uniform.

        An example would be where in a row from a table:

        - multiple columns represent elements of the same feature (e.g. a point represented by x,y,z)
        - the format is not the same for all columns.
        """
        return self.apply(mit.split_into, sizes)

    def split_when(
        self, predicate: Callable[[T, T], bool], max_split: int = -1
    ) -> Iter[list[T]]:
        """
        Split iterable into pieces based on the output of a predicate function.

        Args:
            predicate: Function that takes successive pairs of items and returns True if the iterable should be split.
            max_split: Maximum number of splits to perform. Defaults to -1 (no limit).

        For example, to find runs of increasing numbers, split the iterable when element i is larger than element i + 1:
        ```python
        >>> import pyochain as pc
        >>> data = pc.Seq([1, 2, 3, 3, 2, 5, 2, 4, 2])
        >>> data.iter().split_when(lambda x, y: x > y).into(list)
        [[1, 2, 3, 3], [2, 5], [2, 4], [2]]

        ```

        At most max_split splits are done.

        If max_split is not specified or -1, then there is no limit on the number of splits:
        ```python
        >>> data.iter().split_when(lambda x, y: x > y, max_split=2).into(list)
        [[1, 2, 3, 3], [2, 5], [2, 4, 2]]

        ```
        """
        return self.apply(mit.split_when, predicate, max_split)

    def chunks(self, n: int, strict: bool = False) -> Iter[list[T]]:
        """
        Break iterable into lists of length n.

        By default, the last yielded list will have fewer than *n* elements if the length of *iterable* is not divisible by *n*.

        To use a fill-in value instead, see the :func:`grouper` recipe.

        If:

        - the length of *iterable* is not divisible by *n*
        - *strict* is `True`

        then `ValueError` will be raised before the last list is yielded.
        Args:
            n: Number of elements in each chunk.
            strict: Whether to raise an error if the last chunk is smaller than n. Defaults to False.
        Example:
        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_([1, 2, 3, 4, 5, 6]).chunks(3).into(list)
        [[1, 2, 3], [4, 5, 6]]
        >>> pc.Iter.from_([1, 2, 3, 4, 5, 6, 7, 8]).chunks(3).into(list)
        [[1, 2, 3], [4, 5, 6], [7, 8]]

        ```
        """
        return self.apply(mit.chunked, n, strict)

    def chunks_even(self, n: int) -> Iter[list[T]]:
        """
        Break iterable into lists of approximately length n.

        Items are distributed such the lengths of the lists differ by at most 1 item.
        Args:
            n: Approximate number of elements in each chunk.
        Example:
        ```python
        >>> import pyochain as pc
        >>> iterable = pc.Seq([1, 2, 3, 4, 5, 6, 7])
        >>> iterable.iter().chunks_even(3).into(list)  # List lengths: 3, 2, 2
        [[1, 2, 3], [4, 5], [6, 7]]
        >>> iterable.iter().chunks(3).into(list)  # List lengths: 3, 3, 1
        [[1, 2, 3], [4, 5, 6], [7]]

        ```
        """
        return self.apply(mit.chunked_even, n)

    def unique_to_each[U: Iterable[Any]](self: IterWrapper[U]) -> Iter[list[U]]:
        """
        Return the elements from each of the iterables that aren't in the other iterables.

        For example, suppose you have a set of packages, each with a set of dependencies:

        **{'pkg_1': {'A', 'B'}, 'pkg_2': {'B', 'C'}, 'pkg_3': {'B', 'D'}}**

        If you remove one package, which dependencies can also be removed?

        If pkg_1 is removed, then A is no longer necessary - it is not associated with pkg_2 or pkg_3.

        Similarly, C is only needed for pkg_2, and D is only needed for pkg_3:
        ```python
        >>> import pyochain as pc
        >>> data = ({"A", "B"}, {"B", "C"}, {"B", "D"})
        >>> pc.Iter.from_(data).unique_to_each().collect().unwrap()
        [['A'], ['C'], ['D']]

        ```

        If there are duplicates in one input iterable that aren't in the others they will be duplicated in the output.

        Input order is preserved:
        ```python
        >>> data = ("mississippi", "missouri")
        >>> pc.Iter.from_(data).unique_to_each().collect().unwrap()
        [['p', 'p'], ['o', 'u', 'r']]

        ```

        It is assumed that the elements of each iterable are hashable.
        """

        from collections import Counter

        def _unique_to_each(data: Iterable[U]) -> Generator[list[U], None, None]:
            """from more_itertools.unique_to_each"""
            pool: list[Iterable[U]] = [it for it in data]
            counts: Counter[U] = Counter(itertools.chain.from_iterable(map(set, pool)))
            uniques: set[U] = {element for element in counts if counts[element] == 1}
            return ((list(filter(uniques.__contains__, it))) for it in pool)

        return self.apply(_unique_to_each)



from __future__ import annotations

from collections.abc import Callable, Iterable, Mapping
from typing import TYPE_CHECKING, Concatenate

import cytoolz as cz

from .._core import MappingWrapper

if TYPE_CHECKING:
    from .._iter import Iter
    from ._main import Dict


class IterDict[K, V](MappingWrapper[K, V]):
    def itr[**P, R, U](
        self: MappingWrapper[K, Iterable[U]],
        func: Callable[Concatenate[Iter[U], P], R],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> Dict[K, R]:
        """
        Apply a function to each value after wrapping it in an Iter.

        Args:
            func: Function to apply to each value after wrapping it in an Iter.
            *args: Positional arguments to pass to the function.
            **kwargs: Keyword arguments to pass to the function.

        Syntactic sugar for `map_values(lambda data: func(Iter(data), *args, **kwargs))`
        ```python
        >>> import pyochain as pc
        >>> data = {
        ...     "numbers1": [1, 2, 3],
        ...     "numbers2": [4, 5, 6],
        ... }
        >>> pc.Dict(data).itr(lambda v: v.repeat(5).flatten().sum()).unwrap()
        {'numbers1': 30, 'numbers2': 75}

        ```
        """
        from .._iter import Iter

        def _itr(data: Mapping[K, Iterable[U]]) -> dict[K, R]:
            def _(v: Iterable[U]) -> R:
                return func(Iter.from_(v), *args, **kwargs)

            return cz.dicttoolz.valmap(_, data)

        return self.apply(_itr)

    def iter_keys(self) -> Iter[K]:
        """
        Return a Iter of the dict's keys.
        ```python
        >>> import pyochain as pc
        >>> pc.Dict({1: 2}).iter_keys().into(list)
        [1]

        ```
        """
        from .._iter import Iter

        return Iter.from_(self.unwrap().keys())

    def iter_values(self) -> Iter[V]:
        """
        Return an Iter of the dict's values.
        ```python
        >>> import pyochain as pc
        >>> pc.Dict({1: 2}).iter_values().into(list)
        [2]

        ```
        """
        from .._iter import Iter

        return Iter.from_(self.unwrap().values())

    def iter_items(self) -> Iter[tuple[K, V]]:
        """
        Return a Iter of the dict's items.
        ```python
        >>> import pyochain as pc
        >>> pc.Dict({1: 2}).iter_items().into(list)
        [(1, 2)]

        ```
        """
        from .._iter import Iter

        return Iter.from_(self.unwrap().items())



from __future__ import annotations

import itertools
from collections.abc import Callable
from functools import partial
from typing import TYPE_CHECKING, Literal, overload

import cytoolz as cz

from .._core import IterWrapper

if TYPE_CHECKING:
    from ._main import Iter


class BasePartitions[T](IterWrapper[T]):
    @overload
    def windows(self, length: Literal[1]) -> Iter[tuple[T]]: ...
    @overload
    def windows(self, length: Literal[2]) -> Iter[tuple[T, T]]: ...
    @overload
    def windows(self, length: Literal[3]) -> Iter[tuple[T, T, T]]: ...
    @overload
    def windows(self, length: Literal[4]) -> Iter[tuple[T, T, T, T]]: ...
    @overload
    def windows(self, length: Literal[5]) -> Iter[tuple[T, T, T, T, T]]: ...

    def windows(self, length: int) -> Iter[tuple[T, ...]]:
        """
        A sequence of overlapping subsequences of the given length.

        Args:
            length: The length of each window.

        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_([1, 2, 3, 4]).windows(2).into(list)
        [(1, 2), (2, 3), (3, 4)]

        ```
        This function allows you to apply custom function not available in the rolling namespace.
        ```python
        >>> def moving_average(seq: tuple[int, ...]) -> float:
        ...     return float(sum(seq)) / len(seq)
        >>> pc.Iter.from_([1, 2, 3, 4]).windows(2).map(moving_average).into(list)
        [1.5, 2.5, 3.5]

        ```
        """
        return self.apply(partial(cz.itertoolz.sliding_window, length))

    @overload
    def partition(self, n: Literal[1], pad: None = None) -> Iter[tuple[T]]: ...
    @overload
    def partition(self, n: Literal[2], pad: None = None) -> Iter[tuple[T, T]]: ...
    @overload
    def partition(self, n: Literal[3], pad: None = None) -> Iter[tuple[T, T, T]]: ...
    @overload
    def partition(self, n: Literal[4], pad: None = None) -> Iter[tuple[T, T, T, T]]: ...
    @overload
    def partition(
        self, n: Literal[5], pad: None = None
    ) -> Iter[tuple[T, T, T, T, T]]: ...
    @overload
    def partition(self, n: int, pad: int) -> Iter[tuple[T, ...]]: ...
    def partition(self, n: int, pad: int | None = None) -> Iter[tuple[T, ...]]:
        """
        Partition sequence into tuples of length n.

        Args:
            n: Length of each partition.
            pad: Value to pad the last partition if needed.

        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_([1, 2, 3, 4]).partition(2).into(list)
        [(1, 2), (3, 4)]

        ```
        If the length of seq is not evenly divisible by n, the final tuple is dropped if pad is not specified, or filled to length n by pad:
        ```python
        >>> pc.Iter.from_([1, 2, 3, 4, 5]).partition(2).into(list)
        [(1, 2), (3, 4), (5, None)]

        ```
        """

        return self.apply(partial(cz.itertoolz.partition, n, pad=pad))

    def partition_all(self, n: int) -> Iter[tuple[T, ...]]:
        """
        Partition all elements of sequence into tuples of length at most n.

        Args:
            n: Maximum length of each partition.

        The final tuple may be shorter to accommodate extra elements.
        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_([1, 2, 3, 4]).partition_all(2).into(list)
        [(1, 2), (3, 4)]
        >>> pc.Iter.from_([1, 2, 3, 4, 5]).partition_all(2).into(list)
        [(1, 2), (3, 4), (5,)]

        ```
        """
        return self.apply(partial(cz.itertoolz.partition_all, n))

    def partition_by(self, predicate: Callable[[T], bool]) -> Iter[tuple[T, ...]]:
        """
        Partition the `iterable` into a sequence of `tuples` according to a predicate function.

        Args:
            predicate: Function to determine partition boundaries.

        Every time the output of `predicate` changes, a new `tuple` is started,
        and subsequent items are collected into that `tuple`.
        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_("I have space").partition_by(lambda c: c == " ").into(list)
        [('I',), (' ',), ('h', 'a', 'v', 'e'), (' ',), ('s', 'p', 'a', 'c', 'e')]
        >>>
        >>> data = [1, 2, 1, 99, 88, 33, 99, -1, 5]
        >>> pc.Iter.from_(data).partition_by(lambda x: x > 10).into(list)
        [(1, 2, 1), (99, 88, 33, 99), (-1, 5)]

        ```
        """
        return self.apply(partial(cz.recipes.partitionby, predicate))

    def batch(self, n: int) -> Iter[tuple[T, ...]]:
        """
        Batch elements into tuples of length n and return a new Iter.

        Args:
            n: Number of elements in each batch.

        - The last batch may be shorter than n.
        - The data is consumed lazily, just enough to fill a batch.
        - The result is yielded as soon as a batch is full or when the input iterable is exhausted.
        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_("ABCDEFG").batch(3).into(list)
        [('A', 'B', 'C'), ('D', 'E', 'F'), ('G',)]

        ```
        """
        return self.apply(itertools.batched, n)



from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Callable, Collection, Iterable, Iterator
from typing import TYPE_CHECKING, Any, Concatenate, Self

if TYPE_CHECKING:
    from .._dict import Dict
    from .._iter import Iter, Seq


class Pipeable:
    def pipe[**P, R](
        self,
        func: Callable[Concatenate[Self, P], R],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> R:
        """Pipe the instance in the function and return the result."""
        return func(self, *args, **kwargs)


class CommonBase[T](ABC, Pipeable):
    """
    Base class for all wrappers.
    You can subclass this to create your own wrapper types.
    The pipe unwrap method must be implemented to allow piping functions that transform the underlying data type, whilst retaining the wrapper.
    """

    _data: T

    __slots__ = ("_data",)

    def __init__(self, data: T) -> None:
        self._data = data

    @abstractmethod
    def apply[**P](
        self,
        func: Callable[Concatenate[T, P], Any],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> Any:
        raise NotImplementedError

    def println(self, pretty: bool = True) -> Self:
        """
        Print the underlying data and return self for chaining.

        Useful for debugging, simply insert `.println()` in the chain,
        and then removing it will not affect the rest of the chain.
        """
        from pprint import pprint

        if pretty:
            pprint(self.unwrap(), sort_dicts=False)
        else:
            print(self.unwrap())
        return self

    def unwrap(self) -> T:
        """
        Return the underlying data.

        This is a terminal operation.
        """
        return self._data

    def into[**P, R](
        self,
        func: Callable[Concatenate[T, P], R],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> R:
        """
        Pass the *unwrapped* underlying data into a function.

        The result is not wrapped.
        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_(range(5)).into(list)
        [0, 1, 2, 3, 4]

        ```
        This is a core functionality that allows ending the chain whilst keeping the code style consistent.
        """
        return func(self.unwrap(), *args, **kwargs)


class IterWrapper[T](CommonBase[Iterable[T]]):
    _data: Iterable[T]

    def apply[**P, R](
        self,
        func: Callable[Concatenate[Iterable[T], P], Iterator[R]],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> Iter[R]:
        """
        Apply a function to the underlying iterable and return an Iter of the result.
        Allow to pass user defined functions that transform the iterable while retaining the Iter wrapper.
        Args:
            func: Function to apply to the underlying iterable.
            *args: Positional arguments to pass to the function.
            **kwargs: Keyword arguments to pass to the function.

        Example:
        ```python
        >>> import pyochain as pc
        >>> def double(data: Iterable[int]) -> Iterator[int]:
        ...     return (x * 2 for x in data)
        >>> pc.Iter.from_([1, 2, 3]).apply(double).into(list)
        [2, 4, 6]
        """
        from .._iter import Iter

        return Iter(self.into(func, *args, **kwargs))

    def collect(self, factory: Callable[[Iterable[T]], Collection[T]] = list) -> Seq[T]:
        """
        Collect the elements into a sequence.
        Args:
            factory: A callable that takes an iterable and returns a collection. Defaults to list.

        Example:
        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_(range(5)).collect().unwrap()
        [0, 1, 2, 3, 4]

        ```
        """
        from .._iter import Seq

        return Seq(self.into(factory))


class MappingWrapper[K, V](CommonBase[dict[K, V]]):
    _data: dict[K, V]

    def apply[**P, KU, VU](
        self,
        func: Callable[Concatenate[dict[K, V], P], dict[KU, VU]],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> Dict[KU, VU]:
        """
        Apply a function to the underlying dict and return a Dict of the result.
        Allow to pass user defined functions that transform the dict while retaining the Dict wrapper.
        Args:
            func: Function to apply to the underlying dict.
            *args: Positional arguments to pass to the function.
            **kwargs: Keyword arguments to pass to the function.
        Example:
        ```python
        >>> import pyochain as pc
        >>> def invert_dict(d: dict[K, V]) -> dict[V, K]:
        ...     return {v: k for k, v in d.items()}
        >>> pc.Dict({'a': 1, 'b': 2}).apply(invert_dict).unwrap()
        {1: 'a', 2: 'b'}

        ```
        """
        from .._dict import Dict

        return Dict(self.into(func, *args, **kwargs))


class Wrapper[T](CommonBase[T]):
    """
    A generic Wrapper for any type.
    The pipe into method is implemented to return a Wrapper of the result type.

    This class is intended for use with other types/implementations that do not support the fluent/functional style.
    This allow the use of a consistent code style across the code base.
    """

    def apply[**P, R](
        self,
        func: Callable[Concatenate[T, P], R],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> Wrapper[R]:
        return Wrapper(self.into(func, *args, **kwargs))



from __future__ import annotations

from collections.abc import Callable, Mapping
from functools import partial
from typing import TYPE_CHECKING, Any, TypeGuard

import cytoolz as cz

from .._core import MappingWrapper

if TYPE_CHECKING:
    from ._main import Dict


class FilterDict[K, V](MappingWrapper[K, V]):
    def filter_keys(self, predicate: Callable[[K], bool]) -> Dict[K, V]:
        """
        Return a new Dict containing keys that satisfy predicate.

        Args:
            predicate: Function to determine if a key should be included.
        Example:
        ```python
        >>> import pyochain as pc
        >>> d = {1: 2, 2: 3, 3: 4, 4: 5}
        >>> pc.Dict(d).filter_keys(lambda x: x % 2 == 0).unwrap()
        {2: 3, 4: 5}

        ```
        """
        return self.apply(partial(cz.dicttoolz.keyfilter, predicate))

    def filter_values(self, predicate: Callable[[V], bool]) -> Dict[K, V]:
        """
        Return a new Dict containing items whose values satisfy predicate.

        Args:
            predicate: Function to determine if a value should be included.
        Example:
        ```python
        >>> import pyochain as pc
        >>> d = {1: 2, 2: 3, 3: 4, 4: 5}
        >>> pc.Dict(d).filter_values(lambda x: x % 2 == 0).unwrap()
        {1: 2, 3: 4}
        >>> pc.Dict(d).filter_values(lambda x: not x > 3).unwrap()
        {1: 2, 2: 3}

        ```
        """
        return self.apply(partial(cz.dicttoolz.valfilter, predicate))

    def filter_items(
        self,
        predicate: Callable[[tuple[K, V]], bool],
    ) -> Dict[K, V]:
        """
        Filter items by predicate applied to (key, value) tuples.

        Args:
            predicate: Function to determine if a (key, value) pair should be included.
        Example:
        ```python
        >>> import pyochain as pc
        >>> def isvalid(item):
        ...     k, v = item
        ...     return k % 2 == 0 and v < 4
        >>> d = pc.Dict({1: 2, 2: 3, 3: 4, 4: 5})
        >>>
        >>> d.filter_items(isvalid).unwrap()
        {2: 3}
        >>> d.filter_items(lambda kv: not isvalid(kv)).unwrap()
        {1: 2, 3: 4, 4: 5}

        ```
        """
        return self.apply(partial(cz.dicttoolz.itemfilter, predicate))

    def filter_kv(
        self,
        predicate: Callable[[K, V], bool],
    ) -> Dict[K, V]:
        """
        Filter items by predicate applied to unpacked (key, value) tuples.

        Args:
            predicate: Function to determine if a key-value pair should be included.
        Example:
        ```python
        >>> import pyochain as pc
        >>> def isvalid(key, value):
        ...     return key % 2 == 0 and value < 4
        >>> d = pc.Dict({1: 2, 2: 3, 3: 4, 4: 5})
        >>>
        >>> d.filter_kv(isvalid).unwrap()
        {2: 3}
        >>> d.filter_kv(lambda k, v: not isvalid(k, v)).unwrap()
        {1: 2, 3: 4, 4: 5}

        ```
        """

        def _filter_kv(data: dict[K, V]) -> dict[K, V]:
            def _(kv: tuple[K, V]) -> bool:
                return predicate(kv[0], kv[1])

            return cz.dicttoolz.itemfilter(_, data)

        return self.apply(_filter_kv)

    def filter_attr[U](self, attr: str, dtype: type[U] = object) -> Dict[K, U]:
        """
        Filter values that have a given attribute.

        This does not enforce type checking at runtime for performance considerations.
        Args:
            attr: Attribute name to check for.
            dtype: Optional expected type of the attribute for type hinting.
        Example:
        ```python
        >>> import pyochain as pc
        >>> pc.Dict({"a": "hello", "b": "world", "c": 2, "d": 5}).filter_attr(
        ...     "capitalize", str
        ... ).unwrap()
        {'a': 'hello', 'b': 'world'}

        ```
        """

        def _filter_attr(data: dict[K, V]) -> dict[K, U]:
            def has_attr(x: V) -> TypeGuard[U]:
                return hasattr(x, attr)

            return cz.dicttoolz.valfilter(has_attr, data)

        return self.apply(_filter_attr)

    def filter_type[R](self, typ: type[R]) -> Dict[K, R]:
        """
        Filter values by type.

        Args:
            typ: Type to filter values by.
        Example:
        ```python
        >>> import pyochain as pc
        >>> data = {"a": "one", "b": "two", "c": 3, "d": 4}
        >>> pc.Dict(data).filter_type(str).unwrap()
        {'a': 'one', 'b': 'two'}

        ```
        """

        def _filter_type(data: dict[K, V]) -> dict[K, R]:
            def _(x: V) -> TypeGuard[R]:
                return isinstance(x, typ)

            return cz.dicttoolz.valfilter(_, data)

        return self.apply(_filter_type)

    def filter_callable(self) -> Dict[K, Callable[..., Any]]:
        """
        Filter values that are callable.
        ```python
        >>> import pyochain as pc
        >>> def foo():
        ...     pass
        >>> data = {1: "one", 2: "two", 3: foo, 4: print}
        >>> pc.Dict(data).filter_callable().map_values(lambda x: x.__name__).unwrap()
        {3: 'foo', 4: 'print'}

        ```
        """

        def _filter_callable(data: dict[K, V]) -> dict[K, Callable[..., Any]]:
            def _(x: V) -> TypeGuard[Callable[..., Any]]:
                return callable(x)

            return cz.dicttoolz.valfilter(_, data)

        return self.apply(_filter_callable)

    def filter_subclass[U: type[Any], R](
        self: FilterDict[K, U], parent: type[R], keep_parent: bool = True
    ) -> Dict[K, type[R]]:
        """
        Filter values that are subclasses of a given parent class.

        Args:
            parent: Parent class to check against.
            keep_parent: Whether to include the parent class itself. Defaults to True.

        ```python
        >>> import pyochain as pc
        >>> class A:
        ...     pass
        >>> class B(A):
        ...     pass
        >>> class C:
        ...     pass
        >>> def name(cls: type[Any]) -> str:
        ...     return cls.__name__
        >>> data = pc.Dict({"first": A, "second": B, "third": C})
        >>> data.filter_subclass(A).map_values(name).unwrap()
        {'first': 'A', 'second': 'B'}
        >>> data.filter_subclass(A, keep_parent=False).map_values(name).unwrap()
        {'second': 'B'}

        ```
        """

        def _filter_subclass(data: dict[K, U]) -> dict[K, type[R]]:
            def _(x: type[Any]) -> TypeGuard[type[R]]:
                if keep_parent:
                    return issubclass(x, parent)
                else:
                    return issubclass(x, parent) and x is not parent

            return cz.dicttoolz.valfilter(_, data)

        return self.apply(_filter_subclass)

    def intersect_keys(self, *others: Mapping[K, V]) -> Dict[K, V]:
        """
        Return a new Dict keeping only keys present in self and all others.

        Args:
            *others: Other mappings to intersect keys with.

        ```python
        >>> import pyochain as pc
        >>> d1 = {"a": 1, "b": 2, "c": 3}
        >>> d2 = {"b": 10, "c": 20}
        >>> d3 = {"c": 30}
        >>> pc.Dict(d1).intersect_keys(d2, d3).unwrap()
        {'c': 3}

        ```
        """

        def _intersect_keys(data: dict[K, V]) -> dict[K, V]:
            self_keys = set(data.keys())
            for other in others:
                self_keys.intersection_update(other.keys())
            return {k: data[k] for k in self_keys}

        return self.apply(_intersect_keys)

    def diff_keys(self, *others: Mapping[K, V]) -> Dict[K, V]:
        """
        Return a new Dict keeping only keys present in self but not in others.

        Args:
            *others: Other mappings to exclude keys from.

        ```python
        >>> import pyochain as pc
        >>> d1 = {"a": 1, "b": 2, "c": 3}
        >>> d2 = {"b": 10, "d": 40}
        >>> d3 = {"c": 30}
        >>> pc.Dict(d1).diff_keys(d2, d3).unwrap()
        {'a': 1}

        ```
        """

        def _diff_keys(data: dict[K, V]) -> dict[K, V]:
            self_keys = set(data.keys())
            for other in others:
                self_keys.difference_update(other.keys())
            return {k: data[k] for k in self_keys}

        return self.apply(_diff_keys)



from __future__ import annotations

import itertools
from collections.abc import Callable, Generator, Iterable, Iterator
from functools import partial
from random import Random
from typing import TYPE_CHECKING, Any

import cytoolz as cz
import more_itertools as mit

from .._core import IterWrapper, Peeked

if TYPE_CHECKING:
    from ._main import Iter


def _too_short(item_count: int):
    return mit.raise_(
        ValueError,
        f"Too few items in iterable (got {item_count})",
    )


def _too_long(item_count: int):
    return mit.raise_(
        ValueError,
        f"Too many items in iterable (got at least {item_count})",
    )


class BaseProcess[T](IterWrapper[T]):
    def cycle(self) -> Iter[T]:
        """
        Repeat the sequence indefinitely.

        **Warning** ⚠️
            This creates an infinite iterator.
            Be sure to use Iter.take() or Iter.slice() to limit the number of items taken.
        ```python

        Example:
        >>> import pyochain as pc
        >>> pc.Iter.from_([1, 2]).cycle().take(5).into(list)
        [1, 2, 1, 2, 1]

        ```
        """
        return self.apply(itertools.cycle)

    def interpose(self, element: T) -> Iter[T]:
        """
        Interpose element between items and return a new Iterable wrapper.

        Args:
            element: The element to interpose between items.

        Example:
        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_([1, 2]).interpose(0).into(list)
        [1, 0, 2]

        ```
        """
        return self.apply(partial(cz.itertoolz.interpose, element))

    def random_sample(
        self, probability: float, state: Random | int | None = None
    ) -> Iter[T]:
        """
        Return elements from a sequence with probability of prob.

        Args:
            probability: The probability of including each element.
            state: Random state or seed for deterministic sampling.

        Returns a lazy iterator of random items from seq.

        random_sample considers each item independently and without replacement.

        See below how the first time it returned 13 items and the next time it returned 6 items.
        ```python
        >>> import pyochain as pc
        >>> data = pc.Seq(list(range(100)))
        >>> data.iter().random_sample(0.1).into(list)  # doctest: +SKIP
        [6, 9, 19, 35, 45, 50, 58, 62, 68, 72, 78, 86, 95]
        >>> data.iter().random_sample(0.1).into(list)  # doctest: +SKIP
        [6, 44, 54, 61, 69, 94]
        ```
        Providing an integer seed for random_state will result in deterministic sampling.

        Given the same seed it will return the same sample every time.
        ```python
        >>> data.iter().random_sample(0.1, state=2016).into(list)
        [7, 9, 19, 25, 30, 32, 34, 48, 59, 60, 81, 98]
        >>> data.iter().random_sample(0.1, state=2016).into(list)
        [7, 9, 19, 25, 30, 32, 34, 48, 59, 60, 81, 98]

        ```
        random_state can also be any object with a method random that returns floats between 0.0 and 1.0 (exclusive).
        ```python
        >>> from random import Random
        >>> randobj = Random(2016)
        >>> data.iter().random_sample(0.1, state=randobj).into(list)
        [7, 9, 19, 25, 30, 32, 34, 48, 59, 60, 81, 98]

        ```
        """

        return self.apply(
            partial(cz.itertoolz.random_sample, probability, random_state=state)
        )

    def accumulate(self, func: Callable[[T, T], T]) -> Iter[T]:
        """
        Return cumulative application of binary op provided by the function.

        Args:
            func: A binary function to apply cumulatively.

        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_([1, 2, 3]).accumulate(lambda a, b: a + b).into(list)
        [1, 3, 6]

        ```
        """
        return self.apply(partial(cz.itertoolz.accumulate, func))

    def insert_left(self, value: T) -> Iter[T]:
        """
        Prepend value to the sequence and return a new Iterable wrapper.

        Args:
            value: The value to prepend.

        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_([2, 3]).insert_left(1).into(list)
        [1, 2, 3]

        ```
        """
        return self.apply(partial(cz.itertoolz.cons, value))

    def peekn(self, n: int) -> Iter[T]:
        """
        Print and return sequence after peeking n items.

        Args:
            n: Number of items to peek.

        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_([1, 2, 3]).peekn(2).into(list)
        Peeked 2 values: (1, 2)
        [1, 2, 3]

        ```
        """

        def _peekn(data: Iterable[T]) -> Iterator[T]:
            peeked = Peeked(*cz.itertoolz.peekn(n, data))
            print(f"Peeked {n} values: {peeked.value}")
            return peeked.sequence

        return self.apply(_peekn)

    def peek(self) -> Iter[T]:
        """
        Print and return sequence after peeking first item.
        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_([1, 2]).peek().into(list)
        Peeked value: 1
        [1, 2]

        ```
        """

        def _peek(data: Iterable[T]) -> Iterator[T]:
            peeked = Peeked(*cz.itertoolz.peek(data))
            print(f"Peeked value: {peeked.value}")
            return peeked.sequence

        return self.apply(_peek)

    def merge_sorted(
        self, *others: Iterable[T], sort_on: Callable[[T], Any] | None = None
    ) -> Iter[T]:
        """
        Merge already-sorted sequences.

        Args:
            others: Other sorted iterables to merge.
            sort_on: Optional key function for sorting.

        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_([1, 3]).merge_sorted([2, 4]).into(list)
        [1, 2, 3, 4]

        ```
        """
        return self.apply(cz.itertoolz.merge_sorted, *others, key=sort_on)

    def interleave(self, *others: Iterable[T]) -> Iter[T]:
        """
        Interleave multiple sequences element-wise.

        Args:
            others: Other iterables to interleave.

        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_([1, 2]).interleave([3, 4]).into(list)
        [1, 3, 2, 4]

        ```
        """

        def _interleave(data: Iterable[T]) -> Iterator[T]:
            return cz.itertoolz.interleave((data, *others))

        return self.apply(_interleave)

    def chain(self, *others: Iterable[T]) -> Iter[T]:
        """
        Concatenate zero or more iterables, any of which may be infinite.

        Args:
            others: Other iterables to concatenate.

        An infinite sequence will prevent the rest of the arguments from being included.

        We use chain.from_iterable rather than chain(*seqs) so that seqs can be a generator.
        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_([1, 2]).chain([3, 4], [5]).into(list)
        [1, 2, 3, 4, 5]

        ```
        """

        def _chain(data: Iterable[T]) -> Iterator[T]:
            return itertools.chain.from_iterable((data, *others))

        return self.apply(_chain)

    def elements(self) -> Iter[T]:
        """
        Iterator over elements repeating each as many times as its count.

        Note:
            if an element's count has been set to zero or is a negative
            number, elements() will ignore it.
        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_("ABCABC").elements().sort().unwrap()
        ['A', 'A', 'B', 'B', 'C', 'C']

        ```
        Knuth's example for prime factors of 1836:  2**2 * 3**3 * 17**1
        ```python
        >>> import math
        >>> data = [2, 2, 3, 3, 3, 17]
        >>> pc.Iter.from_(data).elements().into(math.prod)
        1836

        ```
        """
        from collections import Counter

        def _elements(data: Iterable[T]) -> Iterator[T]:
            return Counter(data).elements()

        return self.apply(_elements)

    def reverse(self) -> Iter[T]:
        """
        Return a new Iterable wrapper with elements in reverse order.

        The result is a new iterable over the reversed sequence.

        Note:
            This method must consume the entire iterable to perform the reversal.
        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_([1, 2, 3]).reverse().into(list)
        [3, 2, 1]

        ```
        """

        def _reverse(data: Iterable[T]) -> Iterator[T]:
            return reversed(list(data))

        return self.apply(_reverse)

    def is_strictly_n(
        self,
        n: int,
        too_short: Callable[[int], Iterator[T]] | Callable[[int], None] = _too_short,
        too_long: Callable[[int], Iterator[T]] | Callable[[int], None] = _too_long,
    ) -> Iter[T]:
        """
        Validate that *iterable* has exactly *n* items and return them if it does.

        Args:
            n: The exact number of items expected.
            too_short: Function to call if there are too few items.
            too_long: Function to call if there are too many items.

        If it has fewer than *n* items, call function *too_short* with the actual number of items.

        If it has more than *n* items, call function *too_long* with the number `n + 1`.
        ```python
        >>> import pyochain as pc
        >>> iterable = ["a", "b", "c", "d"]
        >>> n = 4
        >>> pc.Iter.from_(iterable).is_strictly_n(n).into(list)
        ['a', 'b', 'c', 'd']

        ```
        Note that the returned iterable must be consumed in order for the check to
        be made.

        By default, *too_short* and *too_long* are functions that raise`ValueError`.
        ```python
        >>> pc.Iter.from_("ab").is_strictly_n(3).into(
        ...     list
        ... )  # doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
        ...
        ValueError: too few items in iterable (got 2)

        >>> pc.Iter.from_("abc").is_strictly_n(2).into(
        ...     list
        ... )  # doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
        ...
        ValueError: too many items in iterable (got at least 3)

        ```
        You can instead supply functions that do something else.

        *too_short* will be called with the number of items in *iterable*.

        *too_long* will be called with `n + 1`.
        ```python
        >>> def too_short(item_count):
        ...     raise RuntimeError
        >>> pc.Iter.from_("abcd").is_strictly_n(6, too_short=too_short).into(list)
        Traceback (most recent call last):
        ...
        RuntimeError
        >>> def too_long(item_count):
        ...     print("The boss is going to hear about this")
        >>> pc.Iter.from_("abcdef").is_strictly_n(4, too_long=too_long).into(list)
        The boss is going to hear about this
        ['a', 'b', 'c', 'd']

        ```
        """

        def strictly_n_(iterable: Iterable[T]) -> Generator[T, Any, None]:
            """from more_itertools.strictly_n"""
            it = iter(iterable)

            sent = 0
            for item in itertools.islice(it, n):
                yield item
                sent += 1

            if sent < n:
                too_short(sent)
                return

            for item in it:
                too_long(n + 1)
                return

        return self.apply(strictly_n_)



from __future__ import annotations

from collections.abc import Callable, Iterable, Mapping
from typing import TYPE_CHECKING

import cytoolz as cz

from .._core import MappingWrapper

if TYPE_CHECKING:
    from ._main import Dict


class JoinsDict[K, V](MappingWrapper[K, V]):
    def inner_join[W](self, other: Mapping[K, W]) -> Dict[K, tuple[V, W]]:
        """
        Performs an inner join with another mapping based on keys.

        Args:
            other: The mapping to join with.

        Only keys present in both mappings are kept.
        ```python
        >>> import pyochain as pc
        >>> d1 = {"a": 1, "b": 2}
        >>> d2 = {"b": 10, "c": 20}
        >>> pc.Dict(d1).inner_join(d2).unwrap()
        {'b': (2, 10)}

        ```
        """

        def _inner_join(data: Mapping[K, V]) -> dict[K, tuple[V, W]]:
            return {k: (v, other[k]) for k, v in data.items() if k in other}

        return self.apply(_inner_join)

    def left_join[W](self, other: Mapping[K, W]) -> Dict[K, tuple[V, W | None]]:
        """
        Performs a left join with another mapping based on keys.

        Args:
            other: The mapping to join with.

        All keys from the left dictionary (self) are kept.
        ```python
        >>> import pyochain as pc
        >>> d1 = {"a": 1, "b": 2}
        >>> d2 = {"b": 10, "c": 20}
        >>> pc.Dict(d1).left_join(d2).unwrap()
        {'a': (1, None), 'b': (2, 10)}

        ```
        """

        def _left_join(data: Mapping[K, V]) -> dict[K, tuple[V, W | None]]:
            return {k: (v, other.get(k)) for k, v in data.items()}

        return self.apply(_left_join)

    def diff(self, other: Mapping[K, V]) -> Dict[K, tuple[V | None, V | None]]:
        """
        Returns a dict of the differences between this dict and another.

        Args:
            other: The mapping to compare against.

        The keys of the returned dict are the keys that are not shared or have different values.
        The values are tuples containing the value from self and the value from other.
        ```python
        >>> import pyochain as pc
        >>> d1 = {"a": 1, "b": 2, "c": 3}
        >>> d2 = {"b": 2, "c": 4, "d": 5}
        >>> pc.Dict(d1).diff(d2).sort().unwrap()
        {'a': (1, None), 'c': (3, 4), 'd': (None, 5)}

        ```
        """

        def _diff(
            data: Mapping[K, V], other: Mapping[K, V]
        ) -> dict[K, tuple[V | None, V | None]]:
            all_keys: set[K] = data.keys() | other.keys()
            diffs: dict[K, tuple[V | None, V | None]] = {}
            for key in all_keys:
                self_val = data.get(key)
                other_val = other.get(key)
                if self_val != other_val:
                    diffs[key] = (self_val, other_val)
            return diffs

        return self.apply(_diff, other)

    def merge(self, *others: Mapping[K, V]) -> Dict[K, V]:
        """
        Merge other dicts into this one and return a new Dict.

        Args:
            *others: One or more mappings to merge into the current dictionary.

        ```python
        >>> import pyochain as pc
        >>> pc.Dict({1: "one"}).merge({2: "two"}).unwrap()
        {1: 'one', 2: 'two'}
        >>> # Later dictionaries have precedence
        >>> pc.Dict({1: 2, 3: 4}).merge({3: 3, 4: 4}).unwrap()
        {1: 2, 3: 3, 4: 4}

        ```
        """
        return self.apply(cz.dicttoolz.merge, *others)

    def merge_with(
        self, *others: Mapping[K, V], func: Callable[[Iterable[V]], V]
    ) -> Dict[K, V]:
        """
        Merge dicts using a function to combine values for duplicate keys.

        Args:
            *others: One or more mappings to merge into the current dictionary.
            func: Function to combine values for duplicate keys.

        A key may occur in more than one dict, and all values mapped from the key will be passed to the function as a list, such as func([val1, val2, ...]).
        ```python
        >>> import pyochain as pc
        >>> pc.Dict({1: 1, 2: 2}).merge_with({1: 10, 2: 20}, func=sum).unwrap()
        {1: 11, 2: 22}
        >>> pc.Dict({1: 1, 2: 2}).merge_with({2: 20, 3: 30}, func=max).unwrap()
        {1: 1, 2: 20, 3: 30}

        ```
        """

        def _merge_with(data: Mapping[K, V]) -> dict[K, V]:
            return cz.dicttoolz.merge_with(func, data, *others)

        return self.apply(_merge_with)



from __future__ import annotations

from collections.abc import Callable, Mapping
from typing import TYPE_CHECKING, Any, Concatenate

import cytoolz as cz

from .._core import MappingWrapper

if TYPE_CHECKING:
    from ._main import Dict


class ProcessDict[K, V](MappingWrapper[K, V]):
    def for_each[**P](
        self,
        func: Callable[Concatenate[K, V, P], Any],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> Dict[K, V]:
        """
        Apply a function to each key-value pair in the dict for side effects.

        Args:
            func: Function to apply to each key-value pair.
            *args: Positional arguments to pass to the function.
            **kwargs: Keyword arguments to pass to the function.

        Returns the original Dict unchanged.
        ```python
        >>> import pyochain as pc
        >>> pc.Dict({"a": 1, "b": 2}).for_each(
        ...     lambda k, v: print(f"Key: {k}, Value: {v}")
        ... ).unwrap()
        Key: a, Value: 1
        Key: b, Value: 2
        {'a': 1, 'b': 2}

        ```
        """

        def _for_each(data: dict[K, V]) -> dict[K, V]:
            for k, v in data.items():
                func(k, v, *args, **kwargs)
            return data

        return self.apply(_for_each)

    def update_in(
        self, *keys: K, func: Callable[[V], V], default: V | None = None
    ) -> Dict[K, V]:
        """
        Update value in a (potentially) nested dictionary.

        Args:
            *keys: Sequence of keys representing the nested path to update.
            func: Function to apply to the value at the specified path.
            default: Default value to use if the path does not exist, by default None

        Applies the func to the value at the path specified by keys, returning a new Dict with the updated value.

        If the path does not exist, it will be created with the default value (if provided) before applying func.
        ```python
        >>> import pyochain as pc
        >>> inc = lambda x: x + 1
        >>> pc.Dict({"a": 0}).update_in("a", func=inc).unwrap()
        {'a': 1}
        >>> transaction = {
        ...     "name": "Alice",
        ...     "purchase": {"items": ["Apple", "Orange"], "costs": [0.50, 1.25]},
        ...     "credit card": "5555-1234-1234-1234",
        ... }
        >>> pc.Dict(transaction).update_in("purchase", "costs", func=sum).unwrap()
        {'name': 'Alice', 'purchase': {'items': ['Apple', 'Orange'], 'costs': 1.75}, 'credit card': '5555-1234-1234-1234'}
        >>> # updating a value when k0 is not in d
        >>> pc.Dict({}).update_in(1, 2, 3, func=str, default="bar").unwrap()
        {1: {2: {3: 'bar'}}}
        >>> pc.Dict({1: "foo"}).update_in(2, 3, 4, func=inc, default=0).unwrap()
        {1: 'foo', 2: {3: {4: 1}}}

        ```
        """
        return self.apply(cz.dicttoolz.update_in, keys, func, default=default)

    def with_key(self, key: K, value: V) -> Dict[K, V]:
        """
        Return a new Dict with key set to value.

        Args:
            key: Key to set in the dictionary.
            value: Value to associate with the specified key.

        Does not modify the initial dictionary.
        ```python
        >>> import pyochain as pc
        >>> pc.Dict({"x": 1}).with_key("x", 2).unwrap()
        {'x': 2}
        >>> pc.Dict({"x": 1}).with_key("y", 3).unwrap()
        {'x': 1, 'y': 3}
        >>> pc.Dict({}).with_key("x", 1).unwrap()
        {'x': 1}

        ```
        """
        return self.apply(cz.dicttoolz.assoc, key, value)

    def drop(self, *keys: K) -> Dict[K, V]:
        """
        Return a new Dict with given keys removed.

        Args:
            *keys: Sequence of keys to remove from the dictionary.

        New dict has d[key] deleted for each supplied key.
        ```python
        >>> import pyochain as pc
        >>> pc.Dict({"x": 1, "y": 2}).drop("y").unwrap()
        {'x': 1}
        >>> pc.Dict({"x": 1, "y": 2}).drop("y", "x").unwrap()
        {}
        >>> pc.Dict({"x": 1}).drop("y").unwrap()  # Ignores missing keys
        {'x': 1}
        >>> pc.Dict({1: 2, 3: 4}).drop(1).unwrap()
        {3: 4}

        ```
        """
        return self.apply(cz.dicttoolz.dissoc, *keys)

    def rename(self, mapping: Mapping[K, K]) -> Dict[K, V]:
        """
        Return a new Dict with keys renamed according to the mapping.

        Args:
            mapping: A dictionary mapping old keys to new keys.

        Keys not in the mapping are kept as is.
        ```python
        >>> import pyochain as pc
        >>> d = {"a": 1, "b": 2, "c": 3}
        >>> mapping = {"b": "beta", "c": "gamma"}
        >>> pc.Dict(d).rename(mapping).unwrap()
        {'a': 1, 'beta': 2, 'gamma': 3}

        ```
        """

        def _rename(data: dict[K, V]) -> dict[K, V]:
            return {mapping.get(k, k): v for k, v in data.items()}

        return self.apply(_rename)

    def sort(self, reverse: bool = False) -> Dict[K, V]:
        """
        Sort the dictionary by its keys and return a new Dict.

        Args:
            reverse: Whether to sort in descending order. Defaults to False.

        ```python
        >>> import pyochain as pc
        >>> pc.Dict({"b": 2, "a": 1}).sort().unwrap()
        {'a': 1, 'b': 2}

        ```
        """

        def _sort(data: dict[K, V]) -> dict[K, V]:
            return dict(sorted(data.items(), reverse=reverse))

        return self.apply(_sort)



from __future__ import annotations

from collections.abc import Callable, Collection, Generator, Iterable, Iterator
from typing import TYPE_CHECKING, Any, Concatenate

from ._aggregations import BaseAgg
from ._booleans import BaseBool
from ._constructors import IterConstructors
from ._eager import BaseEager
from ._filters import BaseFilter
from ._groups import BaseGroups
from ._joins import BaseJoins
from ._lists import BaseList
from ._maps import BaseMap
from ._partitions import BasePartitions
from ._process import BaseProcess
from ._rolling import BaseRolling
from ._tuples import BaseTuples

if TYPE_CHECKING:
    from .._dict import Dict


class Iter[T](
    BaseAgg[T],
    BaseBool[T],
    BaseFilter[T],
    BaseProcess[T],
    BaseMap[T],
    BaseRolling[T],
    BaseList[T],
    BaseTuples[T],
    BasePartitions[T],
    BaseJoins[T],
    BaseGroups[T],
    BaseEager[T],
    IterConstructors,
):
    """
    A wrapper around Python's built-in iterable types, providing a rich set of functional programming tools.

    It supports lazy evaluation, allowing for efficient processing of large datasets.

    It is not a collection itself, but a wrapper that provides additional methods for working with iterables.

    It can be constructed from any iterable, including `lists`, `tuples`, `sets`, and `generators`.
    """

    __slots__ = ("_data",)

    def __init__(self, data: Iterator[T] | Generator[T, Any, Any]) -> None:
        self._data = data

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.unwrap().__repr__()})"

    def itr[**P, R, U: Iterable[Any]](
        self: Iter[U],
        func: Callable[Concatenate[Iter[U], P], R],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> Iter[R]:
        """
        Apply a function to each element after wrapping it in an Iter.

        This is a convenience method for the common pattern of mapping a function over an iterable of iterables.

        Args:
            func: Function to apply to each wrapped element.
            *args: Positional arguments to pass to the function.
            **kwargs: Keyword arguments to pass to the function.
        Example:
        ```python
        >>> import pyochain as pc
        >>> data = [
        ...     [1, 2, 3],
        ...     [4, 5],
        ...     [6, 7, 8, 9],
        ... ]
        >>> pc.Iter.from_(data).itr(
        ...     lambda x: x.repeat(2).flatten().reduce(lambda a, b: a + b)
        ... ).into(list)
        [12, 18, 60]

        ```
        """

        def _itr(data: Iterable[U]) -> Generator[R, None, None]:
            return (func(Iter.from_(x), *args, **kwargs) for x in data)

        return self.apply(_itr)

    def struct[**P, R, K, V](
        self: Iter[dict[K, V]],
        func: Callable[Concatenate[Dict[K, V], P], R],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> Iter[R]:
        """
        Apply a function to each element after wrapping it in a Dict.

        This is a convenience method for the common pattern of mapping a function over an iterable of dictionaries.
        Args:
            func: Function to apply to each wrapped dictionary.
            *args: Positional arguments to pass to the function.
            **kwargs: Keyword arguments to pass to the function.
        Example:
        ```python
        >>> from typing import Any
        >>> import pyochain as pc

        >>> data: list[dict[str, Any]] = [
        ...     {"name": "Alice", "age": 30, "city": "New York"},
        ...     {"name": "Bob", "age": 25, "city": "Los Angeles"},
        ...     {"name": "Charlie", "age": 35, "city": "New York"},
        ...     {"name": "David", "age": 40, "city": "Paris"},
        ... ]
        >>>
        >>> def to_title(d: pc.Dict[str, Any]) -> pc.Dict[str, Any]:
        ...     return d.map_keys(lambda k: k.title())
        >>> def is_young(d: pc.Dict[str, Any]) -> bool:
        ...     return d.unwrap().get("Age", 0) < 30
        >>> def set_continent(d: pc.Dict[str, Any], value: str) -> dict[str, Any]:
        ...     return d.with_key("Continent", value).unwrap()
        >>>
        >>> pc.Iter.from_(data).struct(to_title).filter_false(is_young).map(
        ...     lambda d: d.drop("Age").with_key("Continent", "NA")
        ... ).map_if(
        ...     lambda d: d.unwrap().get("City") == "Paris",
        ...     lambda d: set_continent(d, "Europe"),
        ...     lambda d: set_continent(d, "America"),
        ... ).group_by(lambda d: d.get("Continent")).map_values(
        ...     lambda d: pc.Iter.from_(d)
        ...     .struct(lambda d: d.drop("Continent").unwrap())
        ...     .into(list)
        ... )  # doctest: +NORMALIZE_WHITESPACE
        Dict({
        'America': [
            {'Name': 'Alice', 'City': 'New York'},
            {'Name': 'Charlie', 'City': 'New York'}
        ],
        'Europe': [
            {'Name': 'David', 'City': 'Paris'}
        ]
        })

        ```
        """
        from .._dict import Dict

        def _struct(data: Iterable[dict[K, V]]) -> Generator[R, None, None]:
            return (func(Dict(x), *args, **kwargs) for x in data)

        return self.apply(_struct)

    def with_keys[K](self, keys: Iterable[K]) -> Dict[K, T]:
        """
        Create a Dict by zipping the iterable with keys.

        Args:
            keys: Iterable of keys to pair with the values.
        Example:
        ```python
        >>> import pyochain as pc
        >>> keys = ["a", "b", "c"]
        >>> values = [1, 2, 3]
        >>> pc.Iter.from_(values).with_keys(keys).unwrap()
        {'a': 1, 'b': 2, 'c': 3}
        >>> # This is equivalent to:
        >>> pc.Iter.from_(keys).zip(values).pipe(
        ...     lambda x: pc.Dict(x.into(dict)).unwrap()
        ... )
        {'a': 1, 'b': 2, 'c': 3}

        ```
        """
        from .._dict import Dict

        return Dict(dict(zip(keys, self.unwrap())))

    def with_values[V](self, values: Iterable[V]) -> Dict[T, V]:
        """
        Create a Dict by zipping the iterable with values.

        Args:
            values: Iterable of values to pair with the keys.
        Example:
        ```python
        >>> import pyochain as pc
        >>> keys = [1, 2, 3]
        >>> values = ["a", "b", "c"]
        >>> pc.Iter.from_(keys).with_values(values).unwrap()
        {1: 'a', 2: 'b', 3: 'c'}
        >>> # This is equivalent to:
        >>> pc.Iter.from_(keys).zip(values).pipe(
        ...     lambda x: pc.Dict(x.into(dict)).unwrap()
        ... )
        {1: 'a', 2: 'b', 3: 'c'}

        ```
        """
        from .._dict import Dict

        return Dict(dict(zip(self.unwrap(), values)))


class Seq[T](BaseAgg[T], BaseEager[T]):
    """
    pyochain.Seq represent an in memory collection.

    Provides a subset of pyochain.Iter methods with eager evaluation, and is the return type of pyochain.Iter.collect().
    """

    __slots__ = ("_data",)

    def __init__(self, data: Collection[T]) -> None:
        self._data = data

    def iter(self) -> Iter[T]:
        """
        Get an iterator over the sequence.
        Call this to switch to lazy evaluation.
        """
        return Iter.from_(self.unwrap())



from __future__ import annotations

from collections import defaultdict
from collections.abc import Callable, Mapping
from functools import partial
from typing import Any, Self

import cytoolz as cz

from .._core import SupportsKeysAndGetItem
from ._exprs import IntoExpr, compute_exprs
from ._filters import FilterDict
from ._funcs import dict_repr
from ._groups import GroupsDict
from ._iter import IterDict
from ._joins import JoinsDict
from ._nested import NestedDict
from ._process import ProcessDict


class Dict[K, V](
    ProcessDict[K, V],
    IterDict[K, V],
    NestedDict[K, V],
    JoinsDict[K, V],
    FilterDict[K, V],
    GroupsDict[K, V],
):
    """
    Wrapper for Python dictionaries with chainable methods.
    """

    __slots__ = ()

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({dict_repr(self.unwrap())})"

    @staticmethod
    def from_[G, I](data: Mapping[G, I] | SupportsKeysAndGetItem[G, I]) -> Dict[G, I]:
        """
        Create a Dict from a mapping or SupportsKeysAndGetItem.

        Args:
            data: A mapping or object supporting keys and item access to convert into a Dict.

        ```python
        >>> import pyochain as pc
        >>> class MyMapping:
        ...     def __init__(self):
        ...         self._data = {1: "a", 2: "b", 3: "c"}
        ...
        ...     def keys(self):
        ...         return self._data.keys()
        ...
        ...     def __getitem__(self, key):
        ...         return self._data[key]
        >>>
        >>> pc.Dict.from_(MyMapping()).unwrap()
        {1: 'a', 2: 'b', 3: 'c'}

        ```
        """
        return Dict(dict(data))

    @staticmethod
    def from_object(obj: object) -> Dict[str, Any]:
        """
        Create a Dict from an object's __dict__ attribute.

        Args:
            obj: The object whose `__dict__` attribute will be used to create the Dict.

        ```python
        >>> import pyochain as pc
        >>> class Person:
        ...     def __init__(self, name: str, age: int):
        ...         self.name = name
        ...         self.age = age
        >>> person = Person("Alice", 30)
        >>> pc.Dict.from_object(person).unwrap()
        {'name': 'Alice', 'age': 30}

        ```
        """
        return Dict(obj.__dict__)

    def select(self: Dict[str, Any], *exprs: IntoExpr) -> Dict[str, Any]:
        """
        Select and alias fields from the dict based on expressions and/or strings.

        Navigate nested fields using the `pyochain.key` function.

        - Chain `key.key()` calls to access nested fields.
        - Use `key.apply()` to transform values.
        - Use `key.alias()` to rename fields in the resulting dict.

        Args:
            *exprs: Expressions or strings to select and alias fields from the dictionary.

        ```python
        >>> import pyochain as pc
        >>> data = {
        ...     "name": "Alice",
        ...     "age": 30,
        ...     "scores": {"eng": [85, 90, 95], "math": [80, 88, 92]},
        ... }
        >>> scores_expr = pc.key("scores")  # save an expression for reuse
        >>> pc.Dict(data).select(
        ...     pc.key("name").alias("student_name"),
        ...     "age",  # shorthand for pc.key("age")
        ...     scores_expr.key("math").alias("math_scores"),
        ...     scores_expr.key("eng")
        ...     .apply(lambda v: pc.Seq(v).mean())
        ...     .alias("average_eng_score"),
        ... ).unwrap()
        {'student_name': 'Alice', 'age': 30, 'math_scores': [80, 88, 92], 'average_eng_score': 90}

        ```
        """

        def _select(data: dict[str, Any]) -> dict[str, Any]:
            return compute_exprs(exprs, data, {})

        return self.apply(_select)

    def with_fields(self: Dict[str, Any], *exprs: IntoExpr) -> Dict[str, Any]:
        """
        Merge aliased expressions into the root dict (overwrite on collision).

        Args:
            *exprs: Expressions to merge into the root dictionary.

        ```python
        >>> import pyochain as pc
        >>> data = {
        ...     "name": "Alice",
        ...     "age": 30,
        ...     "scores": {"eng": [85, 90, 95], "math": [80, 88, 92]},
        ... }
        >>> scores_expr = pc.key("scores")  # save an expression for reuse
        >>> pc.Dict(data).with_fields(
        ...     scores_expr.key("eng")
        ...     .apply(lambda v: pc.Seq(v).mean())
        ...     .alias("average_eng_score"),
        ... ).unwrap()
        {'name': 'Alice', 'age': 30, 'scores': {'eng': [85, 90, 95], 'math': [80, 88, 92]}, 'average_eng_score': 90}

        ```
        """

        def _with_fields(data: dict[str, Any]) -> dict[str, Any]:
            return compute_exprs(exprs, data, data.copy())

        return self.apply(_with_fields)

    def map_keys[T](self, func: Callable[[K], T]) -> Dict[T, V]:
        """
        Return a Dict with keys transformed by func.

        Args:
            func: Function to apply to each key in the dictionary.

        ```python
        >>> import pyochain as pc
        >>> pc.Dict({"Alice": [20, 15, 30], "Bob": [10, 35]}).map_keys(
        ...     str.lower
        ... ).unwrap()
        {'alice': [20, 15, 30], 'bob': [10, 35]}
        >>>
        >>> pc.Dict({1: "a"}).map_keys(str).unwrap()
        {'1': 'a'}

        ```
        """
        return self.apply(partial(cz.dicttoolz.keymap, func))

    def map_values[T](self, func: Callable[[V], T]) -> Dict[K, T]:
        """
        Return a Dict with values transformed by func.

        Args:
            func: Function to apply to each value in the dictionary.

        ```python
        >>> import pyochain as pc
        >>> pc.Dict({"Alice": [20, 15, 30], "Bob": [10, 35]}).map_values(sum).unwrap()
        {'Alice': 65, 'Bob': 45}
        >>>
        >>> pc.Dict({1: 1}).map_values(lambda v: v + 1).unwrap()
        {1: 2}

        ```
        """
        return self.apply(partial(cz.dicttoolz.valmap, func))

    def map_items[KR, VR](
        self,
        func: Callable[[tuple[K, V]], tuple[KR, VR]],
    ) -> Dict[KR, VR]:
        """
        Transform (key, value) pairs using a function that takes a (key, value) tuple.

        Args:
            func: Function to transform each (key, value) pair into a new (key, value) tuple.

        ```python
        >>> import pyochain as pc
        >>> pc.Dict({"Alice": 10, "Bob": 20}).map_items(
        ...     lambda kv: (kv[0].upper(), kv[1] * 2)
        ... ).unwrap()
        {'ALICE': 20, 'BOB': 40}

        ```
        """
        return self.apply(partial(cz.dicttoolz.itemmap, func))

    def map_kv[KR, VR](
        self,
        func: Callable[[K, V], tuple[KR, VR]],
    ) -> Dict[KR, VR]:
        """
        Transform (key, value) pairs using a function that takes key and value as separate arguments.

        Args:
            func: Function to transform each key and value into a new (key, value) tuple.

        ```python
        >>> import pyochain as pc
        >>> pc.Dict({1: 2}).map_kv(lambda k, v: (k + 1, v * 10)).unwrap()
        {2: 20}

        ```
        """

        def _map_kv(data: dict[K, V]) -> dict[KR, VR]:
            def _(kv: tuple[K, V]) -> tuple[KR, VR]:
                return func(kv[0], kv[1])

            return cz.dicttoolz.itemmap(_, data)

        return self.apply(_map_kv)

    def invert(self) -> Dict[V, list[K]]:
        """
        Invert the dictionary, grouping keys by common (and hashable) values.
        ```python
        >>> import pyochain as pc
        >>> d = {"a": 1, "b": 2, "c": 1}
        >>> pc.Dict(d).invert().unwrap()
        {1: ['a', 'c'], 2: ['b']}

        ```
        """

        def _invert(data: dict[K, V]) -> dict[V, list[K]]:
            inverted: dict[V, list[K]] = defaultdict(list)
            for k, v in data.items():
                inverted[v].append(k)
            return dict(inverted)

        return self.apply(_invert)

    def implode(self) -> Dict[K, list[V]]:
        """
        Nest all the values in lists.
        syntactic sugar for map_values(lambda v: [v])
        ```python
        >>> import pyochain as pc
        >>> pc.Dict({1: 2, 3: 4}).implode().unwrap()
        {1: [2], 3: [4]}

        ```
        """

        def _implode(data: dict[K, V]) -> dict[K, list[V]]:
            def _(v: V) -> list[V]:
                return [v]

            return cz.dicttoolz.valmap(_, data)

        return self.apply(_implode)

    def equals_to(self, other: Self | Mapping[Any, Any]) -> bool:
        """
        Check if two records are equal based on their data.

        Args:
            other: Another Dict or mapping to compare against.

        Example:
        ```python
        >>> import pyochain as pc
        >>> d1 = pc.Dict({"a": 1, "b": 2})
        >>> d2 = pc.Dict({"a": 1, "b": 2})
        >>> d3 = pc.Dict({"a": 1, "b": 3})
        >>> d1.equals_to(d2)
        True
        >>> d1.equals_to(d3)
        False

        ```
        """
        return (
            self.unwrap() == other.unwrap()
            if isinstance(other, Dict)
            else self.unwrap() == other
        )



from __future__ import annotations

from collections.abc import Callable, Iterable
from functools import partial
from typing import TYPE_CHECKING, Any

import cytoolz as cz

from .._core import IterWrapper, SupportsRichComparison

if TYPE_CHECKING:
    from ._main import Seq


class BaseEager[T](IterWrapper[T]):
    def sort[U: SupportsRichComparison[Any]](
        self: BaseEager[U], reverse: bool = False, key: Callable[[U], Any] | None = None
    ) -> Seq[U]:
        """
        Sort the elements of the sequence.

        Note:
            This method must consume the entire iterable to perform the sort.
            The result is a new iterable over the sorted sequence.

        Args:
            reverse: Whether to sort in descending order. Defaults to False.
            key: Function to extract a comparison key from each element. Defaults to None.
        Example:
        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_([3, 1, 2]).sort().into(list)
        [1, 2, 3]

        ```
        """

        def _sort(data: Iterable[U]) -> list[U]:
            return sorted(data, reverse=reverse, key=key)

        return self.collect(_sort)

    def tail(self, n: int) -> Seq[T]:
        """
        Return a tuple of the last n elements.

        Args:
            n: Number of elements to return.
        Example:
        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_([1, 2, 3]).tail(2).unwrap()
        (2, 3)

        ```
        """
        return self.collect(partial(cz.itertoolz.tail, n))

    def top_n(self, n: int, key: Callable[[T], Any] | None = None) -> Seq[T]:
        """
        Return a tuple of the top-n items according to key.

        Args:
            n: Number of top elements to return.
            key: Function to extract a comparison key from each element. Defaults to None.
        Example:
        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_([1, 3, 2]).top_n(2).unwrap()
        (3, 2)

        ```
        """
        return self.collect(partial(cz.itertoolz.topk, n, key=key))

    def union(self, *others: Iterable[T]) -> Seq[T]:
        """
        Return the union of this iterable and 'others'.

        Note:
            This method consumes inner data and removes duplicates.

        Args:
            *others: Other iterables to include in the union.
        Example:
        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_([1, 2, 2]).union([2, 3], [4]).iter().sort().unwrap()
        [1, 2, 3, 4]

        ```
        """

        def _union(data: Iterable[T]) -> set[T]:
            return set(data).union(*others)

        return self.collect(_union)

    def intersection(self, *others: Iterable[T]) -> Seq[T]:
        """
        Return the elements common to this iterable and 'others'.

        Note:
            This method consumes inner data, unsorts it, and removes duplicates.

        Args:
            *others: Other iterables to intersect with.
        Example:
        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_([1, 2, 2]).intersection([2, 3], [2]).unwrap()
        {2}

        ```
        """

        def _intersection(data: Iterable[T]) -> set[T]:
            return set(data).intersection(*others)

        return self.collect(_intersection)

    def diff_unique(self, *others: Iterable[T]) -> Seq[T]:
        """
        Return the difference of this iterable and 'others'.
        (Elements in 'self' but not in 'others').

        Note:
            This method consumes inner data, unsorts it, and removes duplicates.

        Args:
            *others: Other iterables to subtract from this iterable.
        Example:
        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_([1, 2, 2]).diff_unique([2, 3]).unwrap()
        {1}

        ```
        """

        def _difference(data: Iterable[T]) -> set[T]:
            return set(data).difference(*others)

        return self.collect(_difference)

    def diff_symmetric(self, *others: Iterable[T]) -> Seq[T]:
        """
        Return the symmetric difference (XOR) of this iterable and 'others'.

        Note:
            This method consumes inner data, unsorts it, and removes duplicates.

        Args:
            *others: Other iterables to compute the symmetric difference with.
        Example:
        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_([1, 2, 2]).diff_symmetric([2, 3]).iter().sort().unwrap()
        [1, 3]
        >>> pc.Iter.from_([1, 2, 3]).diff_symmetric([3, 4, 5]).iter().sort().unwrap()
        [1, 2, 4, 5]

        ```
        """

        def _symmetric_difference(data: Iterable[T]) -> set[T]:
            return set(data).symmetric_difference(*others)

        return self.collect(_symmetric_difference)

    def most_common(self, n: int | None = None) -> Seq[tuple[T, int]]:
        """
        Return the n most common elements and their counts.

        If n is None, then all elements are returned.

        Args:
            n: Number of most common elements to return. Defaults to None (all elements).

        Example:
        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_([1, 1, 2, 3, 3, 3]).most_common(2).unwrap()
        [(3, 3), (1, 2)]

        ```
        """
        from collections import Counter

        from ._main import Seq

        def _most_common(data: Iterable[T]) -> list[tuple[T, int]]:
            return Counter(data).most_common(n)

        return Seq(self.into(_most_common))



from __future__ import annotations

from collections.abc import Callable, Iterable, Iterator
from functools import partial
from typing import TYPE_CHECKING, Any, overload

import cytoolz as cz
import more_itertools as mit

from .._core import IterWrapper

if TYPE_CHECKING:
    from .._dict import Dict
    from ._main import Iter


class BaseGroups[T](IterWrapper[T]):
    def reduce_by[K](
        self, key: Callable[[T], K], binop: Callable[[T, T], T]
    ) -> Dict[K, T]:
        """
        Perform a simultaneous groupby and reduction.

        Args:
            key: Function to compute the key for grouping.
            binop: Binary operation to reduce the grouped elements.
        Example:
        ```python
        >>> from collections.abc import Iterable
        >>> import pyochain as pc
        >>> from operator import add, mul
        >>>
        >>> def is_even(x: int) -> bool:
        ...     return x % 2 == 0
        >>>
        >>> def group_reduce(data: Iterable[int]) -> int:
        ...     return pc.Iter.from_(data).reduce(add)
        >>>
        >>> data = pc.Seq([1, 2, 3, 4, 5])
        >>> data.iter().reduce_by(is_even, add).unwrap()
        {False: 9, True: 6}
        >>> data.iter().group_by(is_even).map_values(group_reduce).unwrap()
        {False: 9, True: 6}

        ```
        But the former does not build the intermediate groups, allowing it to operate in much less space.

        This makes it suitable for larger datasets that do not fit comfortably in memory

        Simple Examples:
        ```python
        >>> pc.Iter.from_([1, 2, 3, 4, 5]).reduce_by(is_even, add).unwrap()
        {False: 9, True: 6}
        >>> pc.Iter.from_([1, 2, 3, 4, 5]).reduce_by(is_even, mul).unwrap()
        {False: 15, True: 8}

        ```
        """
        from .._dict import Dict

        return Dict(self.into(partial(cz.itertoolz.reduceby, key, binop)))

    def group_by[K](self, on: Callable[[T], K]) -> Dict[K, list[T]]:
        """
        Group elements by key function and return a Dict result.

        Args:
            on: Function to compute the key for grouping.
        Example:
        ```python
        >>> import pyochain as pc
        >>> names = [
        ...     "Alice",
        ...     "Bob",
        ...     "Charlie",
        ...     "Dan",
        ...     "Edith",
        ...     "Frank",
        ... ]
        >>> pc.Iter.from_(names).group_by(len).sort()
        ... # doctest: +NORMALIZE_WHITESPACE
        Dict({
            3: ['Bob', 'Dan'],
            5: ['Alice', 'Edith', 'Frank'],
            7: ['Charlie']
        })
        >>>
        >>> iseven = lambda x: x % 2 == 0
        >>> pc.Iter.from_([1, 2, 3, 4, 5, 6, 7, 8]).group_by(iseven)
        ... # doctest: +NORMALIZE_WHITESPACE
        Dict({
            False: [1, 3, 5, 7],
            True: [2, 4, 6, 8]
        })

        ```
        Non-callable keys imply grouping on a member.
        ```python
        >>> data = [
        ...     {"name": "Alice", "gender": "F"},
        ...     {"name": "Bob", "gender": "M"},
        ...     {"name": "Charlie", "gender": "M"},
        ... ]
        >>> pc.Iter.from_(data).group_by("gender").sort()
        ... # doctest: +NORMALIZE_WHITESPACE
        Dict({
            'F': [
                {'name': 'Alice', 'gender': 'F'}
            ],
            'M': [
                {'name': 'Bob', 'gender': 'M'},
                {'name': 'Charlie', 'gender': 'M'}
            ]
        })

        ```
        """
        from .._dict import Dict

        return Dict(self.into(partial(cz.itertoolz.groupby, on)))

    def frequencies(self) -> Dict[T, int]:
        """
        Find number of occurrences of each value in the iterable.
        ```python
        >>> import pyochain as pc
        >>> data = ["cat", "cat", "ox", "pig", "pig", "cat"]
        >>> pc.Iter.from_(data).frequencies().unwrap()
        {'cat': 3, 'ox': 1, 'pig': 2}

        ```
        """
        from .._dict import Dict

        return Dict(self.into(cz.itertoolz.frequencies))

    def count_by[K](self, key: Callable[[T], K]) -> Dict[K, int]:
        """
        Count elements of a collection by a key function.

        Args:
            key: Function to compute the key for counting.
        Example:
        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_(["cat", "mouse", "dog"]).count_by(len).unwrap()
        {3: 2, 5: 1}
        >>> def iseven(x):
        ...     return x % 2 == 0
        >>> pc.Iter.from_([1, 2, 3]).count_by(iseven).unwrap()
        {False: 2, True: 1}

        ```
        """
        from .._dict import Dict

        return Dict(self.into(partial(cz.recipes.countby, key)))

    @overload
    def group_by_transform(
        self,
        keyfunc: None = None,
        valuefunc: None = None,
        reducefunc: None = None,
    ) -> Iter[tuple[T, Iterator[T]]]: ...
    @overload
    def group_by_transform[U](
        self,
        keyfunc: Callable[[T], U],
        valuefunc: None,
        reducefunc: None,
    ) -> Iter[tuple[U, Iterator[T]]]: ...
    @overload
    def group_by_transform[V](
        self,
        keyfunc: None,
        valuefunc: Callable[[T], V],
        reducefunc: None,
    ) -> Iter[tuple[T, Iterator[V]]]: ...
    @overload
    def group_by_transform[U, V](
        self,
        keyfunc: Callable[[T], U],
        valuefunc: Callable[[T], V],
        reducefunc: None,
    ) -> Iter[tuple[U, Iterator[V]]]: ...
    @overload
    def group_by_transform[W](
        self,
        keyfunc: None,
        valuefunc: None,
        reducefunc: Callable[[Iterator[T]], W],
    ) -> Iter[tuple[T, W]]: ...
    @overload
    def group_by_transform[U, W](
        self,
        keyfunc: Callable[[T], U],
        valuefunc: None,
        reducefunc: Callable[[Iterator[T]], W],
    ) -> Iter[tuple[U, W]]: ...
    @overload
    def group_by_transform[V, W](
        self,
        keyfunc: None,
        valuefunc: Callable[[T], V],
        reducefunc: Callable[[Iterator[V]], W],
    ) -> Iter[tuple[T, W]]: ...
    @overload
    def group_by_transform[U, V, W](
        self,
        keyfunc: Callable[[T], U],
        valuefunc: Callable[[T], V],
        reducefunc: Callable[[Iterator[V]], W],
    ) -> Iter[tuple[U, W]]: ...
    def group_by_transform[U, V](
        self,
        keyfunc: Callable[[T], U] | None = None,
        valuefunc: Callable[[T], V] | None = None,
        reducefunc: Any = None,
    ) -> Iter[tuple[Any, ...]]:
        """
        An extension of itertools.groupby that can apply transformations to the grouped data.

        Args:
            keyfunc: Function to compute the key for grouping. Defaults to None.
            valuefunc: Function to transform individual items after grouping. Defaults to None.
            reducefunc: Function to transform each group of items. Defaults to None.

        Example:
        ```python
        >>> import pyochain as pc
        >>> data = pc.Iter.from_("aAAbBBcCC")
        >>> data.group_by_transform(
        ...     lambda k: k.upper(), lambda v: v.lower(), lambda g: "".join(g)
        ... ).into(list)
        [('A', 'aaa'), ('B', 'bbb'), ('C', 'ccc')]

        ```
        Each optional argument defaults to an identity function if not specified.

        group_by_transform is useful when grouping elements of an iterable using a separate iterable as the key.

        To do this, zip the iterables and pass a keyfunc that extracts the first element and a valuefunc that extracts the second element:

        Note that the order of items in the iterable is significant.

        Only adjacent items are grouped together, so if you don't want any duplicate groups, you should sort the iterable by the key function.

        Example:
        ```python
        >>> from operator import itemgetter
        >>> data = pc.Iter.from_([0, 0, 1, 1, 1, 2, 2, 2, 3])
        >>> data.zip("abcdefghi").group_by_transform(itemgetter(0), itemgetter(1)).map(
        ...     lambda kv: (kv[0], "".join(kv[1]))
        ... ).into(list)
        [(0, 'ab'), (1, 'cde'), (2, 'fgh'), (3, 'i')]

        ```
        """

        def _group_by_transform(data: Iterable[T]) -> Iterator[tuple[Any, ...]]:
            return mit.groupby_transform(data, keyfunc, valuefunc, reducefunc)

        return self.apply(_group_by_transform)



from typing import Any


def dict_repr(
    v: object,
    depth: int = 0,
    max_depth: int = 3,
    max_items: int = 6,
    max_str: int = 80,
    indent: int = 2,
) -> str:
    pad = " " * (depth * indent)
    if depth > max_depth:
        return "…"
    match v:
        case dict():
            items: list[tuple[str, Any]] = list(v.items())  # type: ignore
            shown: list[tuple[str, Any]] = items[:max_items]
            if (
                all(
                    not isinstance(val, dict) and not isinstance(val, list)
                    for _, val in shown
                )
                and len(shown) <= 2
            ):
                body = ", ".join(
                    f"{k!r}: {dict_repr(val, depth + 1)}" for k, val in shown
                )
                if len(items) > max_items:
                    body += ", …"
                return "{" + body + "}"
            lines: list[str] = []
            for k, val in shown:
                lines.append(
                    f"{pad}{' ' * indent}{k!r}: {dict_repr(val, depth + 1, max_depth, max_items, max_str, indent)}"
                )
            if len(items) > max_items:
                lines.append(f"{pad}{' ' * indent}…")
            return "{\n" + ",\n".join(lines) + f"\n{pad}" + "}"

        case list():
            elems: list[Any] = v[:max_items]  # type: ignore
            if (
                all(isinstance(x, (int, float, str, bool, type(None))) for x in elems)
                and len(elems) <= 4
            ):
                body = ", ".join(dict_repr(x, depth + 1) for x in elems)
                if len(v) > max_items:  # type: ignore
                    body += ", …"
                return "[" + body + "]"
            lines = [
                f"{pad}{' ' * indent}{dict_repr(x, depth + 1, max_depth, max_items, max_str, indent)}"
                for x in elems
            ]
            if len(v) > max_items:  # type: ignore
                lines.append(f"{pad}{' ' * indent}…")
            return "[\n" + ",\n".join(lines) + f"\n{pad}" + "]"

        case str():
            return repr(v if len(v) <= max_str else v[:max_str] + "…")
        case _:
            return repr(v)



from collections.abc import Iterable, Iterator, Sized
from typing import NamedTuple, Protocol


class Peeked[T](NamedTuple):
    value: T | tuple[T, ...]
    sequence: Iterator[T]


class SupportsDunderLT[T](Protocol):
    def __lt__(self, other: T, /) -> bool: ...


class SupportsDunderGT[T](Protocol):
    def __gt__(self, other: T, /) -> bool: ...


class SupportsDunderLE[T](Protocol):
    def __le__(self, other: T, /) -> bool: ...


class SupportsDunderGE[T](Protocol):
    def __ge__(self, other: T, /) -> bool: ...


class SupportsKeysAndGetItem[K, V](Protocol):
    def keys(self) -> Iterable[K]: ...
    def __getitem__(self, key: K, /) -> V: ...


class SupportsAllComparisons[T](
    SupportsDunderLT[T],
    SupportsDunderGT[T],
    SupportsDunderLE[T],
    SupportsDunderGE[T],
    Protocol,
): ...


type SupportsRichComparison[T] = SupportsDunderLT[T] | SupportsDunderGT[T]


class SizedIterable[T](Sized, Iterable[T]): ...



from __future__ import annotations

from collections.abc import Callable
from typing import TYPE_CHECKING

import cytoolz as cz

from .._core import MappingWrapper

if TYPE_CHECKING:
    from ._main import Dict


class GroupsDict[K, V](MappingWrapper[K, V]):
    def group_by_value[G](self, func: Callable[[V], G]) -> Dict[G, dict[K, V]]:
        """
        Group dict items into sub-dictionaries based on a function of the value.

        Args:
            func: Function to determine the group for each value.

        ```python
        >>> import pyochain as pc
        >>> d = {"a": 1, "b": 2, "c": 3, "d": 2}
        >>> pc.Dict(d).group_by_value(lambda v: v % 2).unwrap()
        {1: {'a': 1, 'c': 3}, 0: {'b': 2, 'd': 2}}

        ```
        """

        def _group_by_value(data: dict[K, V]) -> dict[G, dict[K, V]]:
            def _(kv: tuple[K, V]) -> G:
                return func(kv[1])

            return cz.dicttoolz.valmap(dict, cz.itertoolz.groupby(_, data.items()))

        return self.apply(_group_by_value)

    def group_by_key[G](self, func: Callable[[K], G]) -> Dict[G, dict[K, V]]:
        """
        Group dict items into sub-dictionaries based on a function of the key.

        Args:
            func: Function to determine the group for each key.

        ```python
        >>> import pyochain as pc
        >>> d = {"user_1": 10, "user_2": 20, "admin_1": 100}
        >>> pc.Dict(d).group_by_key(lambda k: k.split("_")[0]).unwrap()
        {'user': {'user_1': 10, 'user_2': 20}, 'admin': {'admin_1': 100}}

        ```
        """

        def _group_by_key(data: dict[K, V]) -> dict[G, dict[K, V]]:
            def _(kv: tuple[K, V]) -> G:
                return func(kv[0])

            return cz.dicttoolz.valmap(dict, cz.itertoolz.groupby(_, data.items()))

        return self.apply(_group_by_key)

    def group_by_key_agg[G, R](
        self,
        key_func: Callable[[K], G],
        agg_func: Callable[[Dict[K, V]], R],
    ) -> Dict[G, R]:
        """
        Group by key function, then apply aggregation function to each sub-dict.

        Args:
            key_func: Function to determine the group for each key.
            agg_func: Function to aggregate each sub-dictionary.

        This avoids materializing intermediate `Dict` objects if you only need
        an aggregated result for each group.
        ```python
        >>> import pyochain as pc
        >>>
        >>> data = {"user_1": 10, "user_2": 20, "admin_1": 100}
        >>> pc.Dict(data).group_by_key_agg(
        ...     key_func=lambda k: k.split("_")[0],
        ...     agg_func=lambda d: d.iter_values().sum(),
        ... ).unwrap()
        {'user': 30, 'admin': 100}
        >>>
        >>> data_files = {
        ...     "file_a.txt": 100,
        ...     "file_b.log": 20,
        ...     "file_c.txt": 50,
        ...     "file_d.log": 5,
        ... }
        >>>
        >>> def get_stats(sub_dict: pc.Dict[str, int]) -> dict[str, Any]:
        ...     return {
        ...         "count": sub_dict.iter_keys().count(),
        ...         "total_size": sub_dict.iter_values().sum(),
        ...         "max_size": sub_dict.iter_values().max(),
        ...         "files": sub_dict.iter_keys().sort().into(list),
        ...     }
        >>>
        >>> pc.Dict(data_files).group_by_key_agg(
        ...     key_func=lambda k: k.split(".")[-1], agg_func=get_stats
        ... ).sort().unwrap()
        {'log': {'count': 2, 'total_size': 25, 'max_size': 20, 'files': ['file_b.log', 'file_d.log']}, 'txt': {'count': 2, 'total_size': 150, 'max_size': 100, 'files': ['file_a.txt', 'file_c.txt']}}

        ```
        """
        from ._main import Dict

        def _group_by_key_agg(data: dict[K, V]) -> dict[G, R]:
            def _key_func(kv: tuple[K, V]) -> G:
                return key_func(kv[0])

            def _agg_func(items: list[tuple[K, V]]) -> R:
                return agg_func(Dict(dict(items)))

            groups = cz.itertoolz.groupby(_key_func, data.items())
            return cz.dicttoolz.valmap(_agg_func, groups)

        return self.apply(_group_by_key_agg)

    def group_by_value_agg[G, R](
        self,
        value_func: Callable[[V], G],
        agg_func: Callable[[Dict[K, V]], R],
    ) -> Dict[G, R]:
        """
        Group by value function, then apply aggregation function to each sub-dict.

        Args:
            value_func: Function to determine the group for each value.
            agg_func: Function to aggregate each sub-dictionary.

        This avoids materializing intermediate `Dict` objects if you only need
        an aggregated result for each group.
        ```python
        >>> import pyochain as pc
        >>>
        >>> data = {"math": "A", "physics": "B", "english": "A"}
        >>> pc.Dict(data).group_by_value_agg(
        ...     value_func=lambda grade: grade,
        ...     agg_func=lambda d: d.iter_keys().count(),
        ... ).unwrap()
        {'A': 2, 'B': 1}
        >>>
        >>> # --- Exemple 2: Agrégation plus complexe ---
        >>> sales_data = {
        ...     "store_1": "Electronics",
        ...     "store_2": "Groceries",
        ...     "store_3": "Electronics",
        ...     "store_4": "Clothing",
        ... }
        >>>
        >>> # Obtain the first store for each category (after sorting store names)
        >>> pc.Dict(sales_data).group_by_value_agg(
        ...     value_func=lambda category: category,
        ...     agg_func=lambda d: d.iter_keys().sort().first(),
        ... ).sort().unwrap()
        {'Clothing': 'store_4', 'Electronics': 'store_1', 'Groceries': 'store_2'}

        ```
        """
        from ._main import Dict

        def _group_by_value_agg(data: dict[K, V]) -> dict[G, R]:
            def _key_func(kv: tuple[K, V]) -> G:
                return value_func(kv[1])

            def _agg_func(items: list[tuple[K, V]]) -> R:
                return agg_func(Dict(dict(items)))

            groups = cz.itertoolz.groupby(_key_func, data.items())
            return cz.dicttoolz.valmap(_agg_func, groups)

        return self.apply(_group_by_value_agg)



from __future__ import annotations

from collections.abc import Callable, Mapping
from functools import partial
from typing import TYPE_CHECKING, Any, Concatenate

import cytoolz as cz

from .._core import MappingWrapper

if TYPE_CHECKING:
    from .._dict import Dict


class NestedDict[K, V](MappingWrapper[K, V]):
    def struct[**P, R, U: dict[Any, Any]](
        self: NestedDict[K, U],
        func: Callable[Concatenate[Dict[K, U], P], R],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> Dict[K, R]:
        """
        Apply a function to each value after wrapping it in a Dict.

        Args:
            func: Function to apply to each value after wrapping it in a Dict.
            *args: Positional arguments to pass to the function.
            **kwargs: Keyword arguments to pass to the function.

        Syntactic sugar for `map_values(lambda data: func(pc.Dict(data), *args, **kwargs))`
        ```python
        >>> import pyochain as pc
        >>> data = {
        ...     "person1": {"name": "Alice", "age": 30, "city": "New York"},
        ...     "person2": {"name": "Bob", "age": 25, "city": "Los Angeles"},
        ... }
        >>> pc.Dict(data).struct(lambda d: d.map_keys(str.upper).drop("AGE").unwrap())
        ... # doctest: +NORMALIZE_WHITESPACE
        Dict({
            'person1': {'NAME': 'Alice', 'CITY': 'New York'},
            'person2': {'NAME': 'Bob', 'CITY': 'Los Angeles'}
        })

        ```
        """
        from ._main import Dict

        def _struct(data: Mapping[K, U]) -> dict[K, R]:
            def _(v: dict[Any, Any]) -> R:
                return func(Dict(v), *args, **kwargs)

            return cz.dicttoolz.valmap(_, data)

        return self.apply(_struct)

    def flatten(
        self: NestedDict[str, Any], sep: str = ".", max_depth: int | None = None
    ) -> Dict[str, Any]:
        """
        Flatten a nested dictionary, concatenating keys with the specified separator.

        Args:
            sep: Separator to use when concatenating keys
            max_depth: Maximum depth to flatten. If None, flattens completely.
        ```python
        >>> import pyochain as pc
        >>> data = {
        ...     "config": {"params": {"retries": 3, "timeout": 30}, "mode": "fast"},
        ...     "version": 1.0,
        ... }
        >>> pc.Dict(data).flatten().unwrap()
        {'config.params.retries': 3, 'config.params.timeout': 30, 'config.mode': 'fast', 'version': 1.0}
        >>> pc.Dict(data).flatten(sep="_").unwrap()
        {'config_params_retries': 3, 'config_params_timeout': 30, 'config_mode': 'fast', 'version': 1.0}
        >>> pc.Dict(data).flatten(max_depth=1).unwrap()
        {'config.params': {'retries': 3, 'timeout': 30}, 'config.mode': 'fast', 'version': 1.0}

        ```
        """

        def _flatten(
            d: dict[Any, Any], parent_key: str = "", current_depth: int = 1
        ) -> dict[str, Any]:
            items: list[tuple[str, Any]] = []
            for k, v in d.items():
                new_key = parent_key + sep + k if parent_key else k
                if isinstance(v, dict) and (
                    max_depth is None or current_depth < max_depth + 1
                ):
                    items.extend(
                        _flatten(v, new_key, current_depth + 1).items()  # type: ignore
                    )
                else:
                    items.append((new_key, v))  # type: ignore
            return dict(items)

        return self.apply(_flatten)

    def with_nested_key(self, *keys: K, value: V) -> Dict[K, V]:
        """
        Set a nested key path and return a new Dict with new, potentially nested, key value pair.

        Args:
            *keys: Sequence of keys representing the nested path.
            value: Value to set at the specified nested path.
        ```python
        >>> import pyochain as pc
        >>> purchase = {
        ...     "name": "Alice",
        ...     "order": {"items": ["Apple", "Orange"], "costs": [0.50, 1.25]},
        ...     "credit card": "5555-1234-1234-1234",
        ... }
        >>> pc.Dict(purchase).with_nested_key(
        ...     "order", "costs", value=[0.25, 1.00]
        ... ).unwrap()
        {'name': 'Alice', 'order': {'items': ['Apple', 'Orange'], 'costs': [0.25, 1.0]}, 'credit card': '5555-1234-1234-1234'}

        ```
        """
        return self.apply(cz.dicttoolz.assoc_in, keys, value=value)

    def schema(self, max_depth: int = 1) -> Dict[str, Any]:
        """
        Return the schema of the dictionary up to a maximum depth.

        Args:
            max_depth: Maximum depth to inspect. Nested dicts beyond this depth are marked as 'dict'.

        When the max depth is reached, nested dicts are marked as 'dict'.
        For lists, only the first element is inspected.
        ```python
        >>> import pyochain as pc
        >>> # Depth 2: we see up to level2
        >>> data = {
        ...     "level1": {"level2": {"level3": {"key": "value"}}},
        ...     "other_key": 123,
        ...     "list_key": [{"sub_key": "sub_value"}],
        ... }
        >>> pc.Dict(data).schema(max_depth=1).unwrap()
        {'level1': 'dict', 'other_key': 'int', 'list_key': 'list'}
        >>> pc.Dict(data).schema(max_depth=2).unwrap()
        {'level1': {'level2': 'dict'}, 'other_key': 'int', 'list_key': 'dict'}
        >>>
        >>> # Depth 3: we see up to level3
        >>> pc.Dict(data).schema(max_depth=3).unwrap()
        {'level1': {'level2': {'level3': 'dict'}}, 'other_key': 'int', 'list_key': {'sub_key': 'str'}}

        ```
        """

        def _schema(data: dict[Any, Any]) -> Any:
            def _recurse_schema(node: Any, current_depth: int) -> Any:
                if isinstance(node, dict):
                    if current_depth >= max_depth:
                        return "dict"
                    return {
                        k: _recurse_schema(v, current_depth + 1)
                        for k, v in node.items()  # type: ignore
                    }
                elif cz.itertoolz.isiterable(node):
                    if current_depth >= max_depth:
                        return type(node).__name__
                    return _recurse_schema(cz.itertoolz.first(node), current_depth + 1)
                else:
                    return type(node).__name__

            return _recurse_schema(data, 0)

        return self.apply(_schema)

    def pluck[U: str | int](self: NestedDict[U, Any], *keys: str) -> Dict[U, Any]:
        """
        Extract values from nested dictionaries using a sequence of keys.

        Args:
            *keys: Sequence of keys to extract values from the nested dictionaries.
        ```python
        >>> import pyochain as pc
        >>> data = {
        ...     "person1": {"name": "Alice", "age": 30},
        ...     "person2": {"name": "Bob", "age": 25},
        ... }
        >>> pc.Dict(data).pluck("name").unwrap()
        {'person1': 'Alice', 'person2': 'Bob'}

        ```
        """

        getter = partial(cz.dicttoolz.get_in, keys)

        def _pluck(data: Mapping[U, Any]) -> dict[U, Any]:
            return cz.dicttoolz.valmap(getter, data)

        return self.apply(_pluck)

    def get_in(self, *keys: K, default: Any = None) -> Any:
        """
        Retrieve a value from a nested dictionary structure.

        Args:
            *keys: Sequence of keys representing the nested path to retrieve the value.
            default: Default value to return if the keys do not exist.

        ```python
        >>> import pyochain as pc
        >>> data = {"a": {"b": {"c": 1}}}
        >>> pc.Dict(data).get_in("a", "b", "c")
        1
        >>> pc.Dict(data).get_in("a", "x", default="Not Found")
        'Not Found'

        ```
        """

        def _get_in(data: Mapping[K, V]) -> Any:
            return cz.dicttoolz.get_in(keys, data, default)

        return self.into(_get_in)
